# pylint: disable=C0103,C0111,C0302,W0212
import json
import random
import string
import pytest
import numpy as np

import quasardb


def connect(uri):
    return quasardb.Cluster(uri)


def direct_connect(conn, node_uri):
    return conn.node(node_uri)


def config():
    return {"uri":
            {"insecure": "qdb://127.0.0.1:2836",
             "secure": "qdb://127.0.0.1:2838"}}


@pytest.fixture
def qdbd_settings(scope="module"):
    user_key = {}
    cluster_key = ""

    with open('user_private.key', 'r') as user_key_file:
        user_key = json.load(user_key_file)
    with open('cluster_public.key', 'r') as cluster_key_file:
        cluster_key = cluster_key_file.read()
    return {
        "uri": {
            "insecure": "qdb://127.0.0.1:2836",
            "secure": "qdb://127.0.0.1:2838"},
        "security": {
            "user_name": user_key['username'],
            "user_private_key": user_key['secret_key'],
            "cluster_public_key": cluster_key}}


@pytest.fixture
def qdbd_connection(qdbd_settings):
    return connect(qdbd_settings.get("uri").get("insecure"))


@pytest.fixture
def qdbd_secure_connection(qdbd_settings):
    return quasardb.Cluster(
        uri=qdbd_settings.get("uri").get("secure"),
        user_name=qdbd_settings.get("security").get("user_name"),
        user_private_key=qdbd_settings.get("security").get("user_private_key"),
        cluster_public_key=qdbd_settings.get("security").get("cluster_public_key"))


@pytest.fixture
def qdbd_direct_connection(qdbd_settings, qdbd_connection):
    return direct_connect(qdbd_connection, qdbd_settings.get(
        "uri").get("insecure").replace("qdb://", ""))


@pytest.fixture
def qdbd_direct_secure_connection(qdbd_settings, qdbd_secure_connection):
    return direct_connect(
        qdbd_secure_connection,
        qdbd_settings.get("uri").get("secure").replace(
            "qdb://",
            ""))


@pytest.fixture
def random_identifier():
    return ''.join(random.choice(string.ascii_lowercase) for _ in range(16))


@pytest.fixture
def random_string():
    return ''.join(
        random.choice(
            string.ascii_uppercase +
            string.digits) for _ in range(16))


@pytest.fixture
def column_name(random_identifier):
    return random_identifier


@pytest.fixture
def tag_name(random_identifier):
    return random_identifier


@pytest.fixture
def tag_names():
    return sorted([''.join(random.choice(string.ascii_lowercase)
                           for _ in range(16)) for _ in range(10)])


@pytest.fixture
def entry_name(random_string):
    return random_string


@pytest.fixture
def random_blob(random_string):
    return np.random.bytes(16)


@pytest.fixture
def random_integer():
    return random.randint(-1000000000, 1000000000)


@pytest.fixture
def blob_entry(qdbd_connection, entry_name):
    return qdbd_connection.blob(entry_name)


@pytest.fixture
def integer_entry(qdbd_connection, entry_name):
    return qdbd_connection.integer(entry_name)


@pytest.fixture
def table_name(entry_name):
    return entry_name


def _create_table(c, table_name):
    t = c.table(table_name)
    double_col = quasardb.ColumnInfo(quasardb.ColumnType.Double, "the_double")
    blob_col = quasardb.ColumnInfo(quasardb.ColumnType.Blob, "the_blob")
    string_col = quasardb.ColumnInfo(quasardb.ColumnType.String, "the_string")
    int64_col = quasardb.ColumnInfo(quasardb.ColumnType.Int64, "the_int64")
    ts_col = quasardb.ColumnInfo(quasardb.ColumnType.Timestamp, "the_ts")

    t.create([double_col, blob_col, string_col, int64_col, ts_col])
    return t


@pytest.fixture
def table(qdbd_connection, entry_name):
    return _create_table(qdbd_connection, entry_name)


@pytest.fixture
def secure_table(qdbd_secure_connection, entry_name):
    return _create_table(qdbd_secure_connection, entry_name)


@pytest.fixture
def intervals():
    start_time = np.datetime64('2017-01-01', 'ns')
    return [(start_time, start_time + np.timedelta64(1, 's'))]


def create_many_intervals():
    start_time = np.datetime64('2017-01-01', 'ns')
    return np.array([(start_time + np.timedelta64(i, 's'))
                     for i in range(10000)]).astype('datetime64[ns]')


@pytest.fixture
def many_intervals():
    return create_many_intervals()



import pytest
import quasardb


def test_perf_disable(qdbd_connection):
    qdbd_connection.perf().disable()


def test_perf_enable(qdbd_connection):
    qdbd_connection.perf().enable()


def test_perf_enable_secure_conn(qdbd_secure_connection):
    qdbd_secure_connection.perf().enable()


def test_perf_get__when_empty(qdbd_connection):
    profiles = qdbd_connection.perf().get()
    assert len(profiles) == 0


def test_perf_get_after_blob_put(qdbd_connection, entry_name, random_string):
    qdbd_connection.perf().enable()
    qdbd_connection.blob(entry_name).put(random_string)
    qdbd_connection.blob(entry_name).get()

    profiles = qdbd_connection.perf().get()
    assert len(profiles) == 2
    assert profiles[0][0] == "blob.put"
    assert profiles[1][0] == "common.get"


def test_perf_clear_after_blob_put(qdbd_connection, entry_name, random_string):
    qdbd_connection.perf().enable()
    qdbd_connection.blob(entry_name).put(random_string)

    profiles = qdbd_connection.perf().get()
    assert len(profiles) != 0

    qdbd_connection.perf().clear()
    profiles = qdbd_connection.perf().get()
    assert len(profiles) == 0


def test_perf_flamechart(qdbd_connection, entry_name, random_string):
    qdbd_connection.perf().enable()
    qdbd_connection.blob(entry_name).put(random_string)

    x = qdbd_connection.perf().get(flame=True)
    assert len(x) > 0

    x = qdbd_connection.perf().get(flame=True)
    assert len(x) == 0


def test_perf_flamechart_to_file(qdbd_connection, entry_name, random_string):
    qdbd_connection.perf().enable()
    qdbd_connection.blob(entry_name).put(random_string)

    filename='test_perf.out'
    xs = qdbd_connection.perf().get(flame=True, outfile=filename)

    x = "\n".join(xs)
    x = x + "\n"


    with open(filename, 'r') as fp:
        assert x == fp.read()



import re
import copy
import quasardb
import logging
from datetime import datetime

logger = logging.getLogger('quasardb.stats')


stats_prefix  = '$qdb.statistics.'
user_pattern  = re.compile(r'\$qdb.statistics.(.*).uid_([0-9]+)$')
total_pattern = re.compile(r'\$qdb.statistics.(.*)$')


def is_user_stat(s):
    return user_pattern.match(s) is not None

def is_cumulative_stat(s):
    # NOTE(leon): It's quite difficult to express in Python that you don't want any
    # regex to _end_ with uid_[0-9]+, because Python's regex engine doesn't support
    # variable width look-behind.
    #
    # An alternative would be to use the PyPi regex library (for POSIX regexes), but
    # want to stay light on dependencies#
    #
    # As such, we define a 'cumulative' stat as anything that's not a user stat.
    # Simple but effective.
    return user_pattern.match(s) is None

def by_node(conn):
    """
    Returns statistic grouped by node URI.

    Parameters:
    conn: quasardb.Cluster
      Active connection to the QuasarDB cluster
    """
    return {x: of_node(conn, x) for x in conn.endpoints()}


def of_node(conn, uri):
    """
    Returns statistic for a single node.

    Parameters:
    conn: quasardb.Cluster
      Active connection to the QuasarDB cluster

    uri: str
      URI of a node in the cluster, e.g. '127.0.0.1:2836'.
    """

    start = datetime.now()

    dconn = conn.node(uri)
    raw = {
        k: _get_stat(dconn,k) for k in dconn.prefix_get(stats_prefix, 1000)}

    ret = {'by_uid': _by_uid(raw),
           'cumulative': _cumulative(raw)}

    check_duration = datetime.now() - start

    ret['cumulative']['check.online'] = 1
    ret['cumulative']['check.duration_ms'] = int(check_duration.total_seconds() * 1000)

    return ret

_stat_types = {'requests.bytes_out': 'counter',
               'requests.bytes_in': 'counter',
               'async_pipelines.merge.bucket_count': 'counter',
               'async_pipelines.merge.duration_us': 'counter',
               'async_pipelines.write.successes_count': 'counter',
               'async_pipelines.write.failures_count': 'counter',
               'async_pipelines.write.time_us': 'counter'}

async_pipeline_bytes_pattern  = re.compile(r'async_pipelines.pipe_[0-9]+.merge_map.bytes')
async_pipeline_count_pattern  = re.compile(r'async_pipelines.pipe_[0-9]+.merge_map.count')

def _stat_type(stat_id):
    """
    Returns the statistic type by a stat id. Returns one of:

     - 'gauge'
     - 'counter'

    This is useful for determining which value should be reported in a dashboard.
    """
    if stat_id in _stat_types:
        return _stat_types[stat_id]
    elif stat_id.endswith('total_ns'):
        return 'counter'
    elif async_pipeline_bytes_pattern.match(stat_id):
        return 'gauge'
    elif async_pipeline_count_pattern.match(stat_id):
        return 'gauge'
    else:
        return None

def _calculate_delta_stat(stat_id, prev, cur):
    logger.info("calculating delta for stat_id = {}, prev = {}. cur = {}".format(stat_id, prev, cur))

    stat_type = _stat_type(stat_id)
    if stat_type == 'counter':
        return cur - prev
    elif stat_type == 'gauge':
        return cur
    else:
        return None

def _calculate_delta_stats(prev_stats, cur_stats):
    ret = {}
    for stat_id in cur_stats.keys():
        try:
            prev_stat = cur_stats[stat_id]
            cur_stat  = cur_stats[stat_id]

            value = _calculate_delta_stat(stat_id, prev_stat, cur_stat)
            if value is not None:
                ret[stat_id] = value

        except KeyError:
            # Stat likely was not present yet in prev_stats
            pass

    return ret


def calculate_delta(prev, cur):
    """
    Calculates the 'delta' between two successive statistic measurements.
    """
    ret = {}
    for node_id in cur.keys():
        ret[node_id] = _calculate_delta_stats(prev[node_id]['cumulative'],
                                              cur[node_id]['cumulative'])

    return ret

def _get_stat(dconn, k):
    # Ugly, but works: try to retrieve as integer, if not an int, retrieve as
    # blob
    try:
        return dconn.integer(k).get()
    except quasardb.quasardb.Error:
        blob = dconn.blob(k).get()
        return blob.decode('utf-8', 'replace')

def _by_uid(stats):
    xs = {}
    for k, v in stats.items():
        matches = user_pattern.match(k)
        if is_user_stat(k) and matches:
            (metric, uid_str) = matches.groups()
            uid = int(uid_str)
            if uid not in xs:
                xs[uid] = {}

            if not metric.startswith('serialized'):
                xs[uid][metric] = v

    return xs


def _cumulative(stats):
    xs = {}

    for k, v in stats.items():
        matches = total_pattern.match(k)
        if is_cumulative_stat(k) and matches:
            metric = matches.groups()[0]
            if not metric.startswith('serialized'):
                xs[metric] = v

    return xs



import pytest
import quasardb


def test_node_wrong_leading_qdb_in_uri(qdbd_connection, qdbd_settings):
    with pytest.raises(quasardb.Error, match=r'The (handle|argument) is invalid'):
        # With check in connect, node() throws.
        direct_conn = qdbd_connection.node(
            qdbd_settings.get("uri").get("insecure"))
        # Without it, C API should return an error in prefix_get.
        direct_conn.prefix_get("$qdb", 10)


def test_node_empty_prefix(qdbd_direct_connection):
    res = qdbd_direct_connection.prefix_get("testazeazeaze", 10)
    assert len(res) == 0


def test_node_find_prefixes(qdbd_direct_connection):
    res = qdbd_direct_connection.prefix_get("$qdb.statistics.", 20)
    assert len(res) == 20


def test_node_blob_get(qdbd_direct_connection):
    entry = qdbd_direct_connection.blob("$qdb.statistics.node_id")
    got = entry.get()
    assert isinstance(got, bytes) and len(got) > 0


def test_node_integer_get(qdbd_direct_connection):
    entry = qdbd_direct_connection.integer("$qdb.statistics.cpu.system")
    got = entry.get()

    assert isinstance(got, int) and got > 0



# pylint: disable=C0103,C0111,C0302,W0212
import datetime
import pytest
import numpy as np
import quasardb


def _generate_dates(start_time, count):
    return np.array([(start_time + np.timedelta64(i, 's'))
                     for i in range(count)]).astype('datetime64[ns]')


def _generate_double_ts(start_time, count):
    return (_generate_dates(start_time, count),
            np.random.uniform(-100.0, 100.0, count))


def _generate_int64_ts(start_time, count):
    return (_generate_dates(start_time, count),
            np.random.randint(-100, 100, count))


def _generate_timestamp_ts(start_time, start_val, count):
    return (
        _generate_dates(start_time, count),
        _generate_dates(start_val, count))


def _generate_blob_ts(start_time, count):
    dates = _generate_dates(start_time, count)
    values = np.array(
        list(
            np.random.bytes(
                np.random.randint(
                    16,
                    32)) for i in range(count)),
        dtype=np.object_)
    return (dates, values)


def _generate_string_ts(start_time, count):
    dates = _generate_dates(start_time, count)
    values = np.array([("content_" + str(item)) for item in range(count)],
                      dtype=np.unicode)

    return (dates, values)


def _check_ts_results(results, generated, count):
    assert len(results) == 2

    np.testing.assert_array_equal(results[0][:count], generated[0][:count])
    np.testing.assert_array_equal(results[1][:count], generated[1][:count])

def test_column_info_repr(column_name):
    double = quasardb.ColumnInfo(quasardb.ColumnType.Double, column_name)
    blob = quasardb.ColumnInfo(quasardb.ColumnType.Blob, column_name)
    int64 = quasardb.ColumnInfo(quasardb.ColumnType.Int64, column_name)
    timestamp = quasardb.ColumnInfo(quasardb.ColumnType.Timestamp, column_name)
    string = quasardb.ColumnInfo(quasardb.ColumnType.String, column_name)

    assert column_name in str(double)
    assert column_name in str(blob)
    assert column_name in str(int64)
    assert column_name in str(timestamp)
    assert column_name in str(string)

    assert 'double'    in str(double)
    assert 'blob'      in str(blob)
    assert 'int64'     in str(int64)
    assert 'timestamp' in str(timestamp)
    assert 'string'    in str(string)


def test_list_columns_throws_when_timeseries_does_not_exist(
        qdbd_connection, entry_name):
    table = qdbd_connection.table(entry_name)
    with pytest.raises(quasardb.Error):
        table.list_columns()


def test_insert_throws_when_timeseries_does_not_exist(
        qdbd_connection, entry_name):
    table = qdbd_connection.table(entry_name)

    with pytest.raises(quasardb.Error):
        table.double_insert("the_double",
                            np.array(np.datetime64('2011-01-01', 'ns')),
                            np.array([1.0]))


def test_get_ranges_throws_when_timeseries_does_not_exist(
        qdbd_connection, entry_name, intervals):
    table = qdbd_connection.table(entry_name)
    with pytest.raises(quasardb.Error):
        table.double_get_ranges("blah", intervals)


def test_erase_ranges_throw_when_timeseries_does_not_exist(
        qdbd_connection, entry_name, intervals):
    table = qdbd_connection.table(entry_name)
    with pytest.raises(quasardb.Error):
        table.erase_ranges("blah", intervals)


def test_create_without_columns(qdbd_connection, entry_name):
    table = qdbd_connection.table(entry_name)
    table.create([])

    assert len(table.list_columns()) == 0


def test_create_with_shard_size_less_than_1_millisecond_throws(
        qdbd_connection, entry_name):
    table = qdbd_connection.table(entry_name)
    with pytest.raises(quasardb.Error):
        table.create([], datetime.timedelta(milliseconds=0))


def test_create_with_shard_size_of_1_millisecond(qdbd_connection, entry_name):
    table = qdbd_connection.table(entry_name)
    table.create([], datetime.timedelta(milliseconds=1))
    assert len(table.list_columns()) == 0


def test_create_with_shard_size_of_1_day(qdbd_connection, entry_name):
    table = qdbd_connection.table(entry_name)
    table.create([], datetime.timedelta(hours=24))
    assert len(table.list_columns()) == 0


def test_create_with_shard_size_of_4_weeks(qdbd_connection, entry_name):
    table = qdbd_connection.table(entry_name)
    table.create([], datetime.timedelta(weeks=4))
    assert len(table.list_columns()) == 0


def test_create_with_shard_size_of_more_than_1_year(
        qdbd_connection, entry_name):
    table = qdbd_connection.table(entry_name)
    table.create([], datetime.timedelta(weeks=52))
    assert len(table.list_columns()) == 0


def test_table_layout(table):
    col_list = table.list_columns()
    assert len(col_list) == 5

    assert col_list[0].name == "the_double"
    assert col_list[0].type == quasardb.ColumnType.Double

    assert col_list[1].name == "the_blob"
    assert col_list[1].type == quasardb.ColumnType.Blob

    assert col_list[2].name == "the_string"
    assert col_list[2].type == quasardb.ColumnType.String

    assert col_list[3].name == "the_int64"
    assert col_list[3].type == quasardb.ColumnType.Int64

    assert col_list[4].name == "the_ts"
    assert col_list[4].type == quasardb.ColumnType.Timestamp


def test_column_lookup(table):
    assert table.column_index_by_id("the_double") == 0
    assert table.column_index_by_id("the_blob") == 1
    assert table.column_index_by_id("the_string") == 2
    assert table.column_index_by_id("the_int64") == 3
    assert table.column_index_by_id("the_ts") == 4

    with pytest.raises(quasardb.Error):
        table.column_index_by_id('foobar')

    with pytest.raises(TypeError):
        table.column_index_by_id(None)

    with pytest.raises(TypeError):
        table.column_index_by_id()


def test_cannot_double_create(table, entry_name):
    with pytest.raises(quasardb.AliasAlreadyExistsError):
        table.create([quasardb.ColumnInfo(
            quasardb.ColumnType.Double, entry_name)])


def _double_col_name(table):
    return table.list_columns()[0].name


def _blob_col_name(table):
    return table.list_columns()[1].name


def _string_col_name(table):
    return table.list_columns()[2].name


def _int64_col_name(table):
    return table.list_columns()[3].name


def _ts_col_name(table):
    return table.list_columns()[4].name


def _start_time(intervals):
    return intervals[0][0]


def _start_year(intervals):
    return intervals[0][0]


###
# Double tests
#

def test_double_get_ranges__when_timeseries_is_empty(table, intervals):
    results = table.double_get_ranges(_double_col_name(table), intervals)
    assert len(results) == 2
    assert len(results[0]) == 0
    assert len(results[1]) == 0


def test_double_erase_ranges__when_timeseries_is_empty(table, intervals):
    erased_count = table.erase_ranges(_double_col_name(table), intervals)
    assert erased_count == 0


def test_double_get_ranges(table, intervals):
    inserted_double_data = _generate_double_ts(_start_time(intervals), 1000)
    table.double_insert(_double_col_name(table),
                        inserted_double_data[0],
                        inserted_double_data[1])

    results = table.double_get_ranges(_double_col_name(table), [(
        _start_time(intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    _check_ts_results(results, inserted_double_data, 10)

    results = table.double_get_ranges(
        _double_col_name(table),
        [
            (_start_time(intervals),
             _start_time(intervals) + np.timedelta64(10, 's')),
            (_start_time(intervals) + np.timedelta64(10, 's'),
             _start_time(intervals) + np.timedelta64(20, 's'))])

    _check_ts_results(results, inserted_double_data, 20)

    # empty result
    out_of_time = _start_time(intervals) + np.timedelta64(10, 'h')
    results = table.double_get_ranges(_double_col_name(
        table), [(out_of_time, out_of_time + np.timedelta64(10, 's'))])
    assert len(results) == 2
    assert len(results[0]) == 0
    assert len(results[1]) == 0

    # error: column doesn't exist
    with pytest.raises(quasardb.Error):
        table.double_get_ranges("lolilol", [(_start_time(
            intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    with pytest.raises(quasardb.Error):
        table.double_insert(
            "lolilol",
            inserted_double_data[0],
            inserted_double_data[1])

    with pytest.raises(TypeError):
        table.blob_get_ranges(_double_col_name, [(_start_time(
            intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    with pytest.raises(quasardb.IncompatibleTypeError):
        table.blob_insert(
            _double_col_name(table),
            inserted_double_data[0],
            inserted_double_data[1])


def test_double_erase_ranges(table, intervals):
    inserted_double_data = _generate_double_ts(_start_time(intervals), 1000)
    table.double_insert(
        _double_col_name(table),
        inserted_double_data[0],
        inserted_double_data[1])

    results = table.double_get_ranges(_double_col_name(table), [(
        _start_time(intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    erased_count = table.erase_ranges(_double_col_name(table), [(
        _start_time(intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    assert erased_count == len(results[0])

    erased_count = table.erase_ranges(_double_col_name(table), [(
        _start_time(intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    assert erased_count == 0

    results = table.double_get_ranges(_double_col_name(table), [(
        _start_time(intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    assert len(results[0]) == 0


###
# Int64 tests
#

def test_int64_get_ranges__when_timeseries_is_empty(table, intervals):
    results = table.int64_get_ranges(_int64_col_name(table), intervals)
    assert len(results) == 2
    assert len(results[0]) == 0
    assert len(results[1]) == 0


def test_int64_erase_ranges__when_timeseries_is_empty(table, intervals):
    erased_count = table.erase_ranges(_int64_col_name(table), intervals)
    assert erased_count == 0


def test_int64_get_ranges(table, intervals):
    inserted_int64_data = _generate_int64_ts(_start_time(intervals), 1000)
    table.int64_insert(_int64_col_name(table),
                       inserted_int64_data[0],
                       inserted_int64_data[1])

    results = table.int64_get_ranges(_int64_col_name(table), [(
        _start_time(intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    _check_ts_results(results, inserted_int64_data, 10)

    results = table.int64_get_ranges(_int64_col_name(table),
                                     [(_start_time(intervals),
                                       _start_time(intervals) + np.timedelta64(10, 's')),
                                      (_start_time(intervals) + np.timedelta64(10, 's'),
                                       _start_time(intervals) + np.timedelta64(20, 's'))])

    _check_ts_results(results, inserted_int64_data, 20)

    # Everything
    results = table.int64_get_ranges(_int64_col_name(table))
    _check_ts_results(results, inserted_int64_data, 1000)

    # empty result
    out_of_time = _start_time(intervals) + np.timedelta64(10, 'h')
    results = table.int64_get_ranges(_int64_col_name(
        table), [(out_of_time, out_of_time + np.timedelta64(10, 's'))])
    assert len(results) == 2
    assert len(results[0]) == 0
    assert len(results[1]) == 0

    # error: column doesn't exist
    with pytest.raises(quasardb.Error):
        table.int64_get_ranges("lolilol", [(_start_time(
            intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    with pytest.raises(quasardb.Error):
        table.int64_insert(
            "lolilol",
            inserted_int64_data[0],
            inserted_int64_data[1])

    with pytest.raises(TypeError):
        table.blob_get_ranges(_int64_col_name, [(_start_time(
            intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    with pytest.raises(quasardb.IncompatibleTypeError):
        table.blob_insert(
            _int64_col_name(table),
            inserted_int64_data[0],
            inserted_int64_data[1])


def test_int64_erase_ranges(table, intervals):
    inserted_int64_data = _generate_int64_ts(_start_time(intervals), 1000)
    table.int64_insert(
        _int64_col_name(table),
        inserted_int64_data[0],
        inserted_int64_data[1])

    results = table.int64_get_ranges(_int64_col_name(table), [(
        _start_time(intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    erased_count = table.erase_ranges(_int64_col_name(table), [(
        _start_time(intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    assert erased_count == len(results[0])

    erased_count = table.erase_ranges(_int64_col_name(table), [(
        _start_time(intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    assert erased_count == 0

    results = table.int64_get_ranges(_int64_col_name(table), [(
        _start_time(intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    assert len(results[0]) == 0


###
# Blob tests
#

def test_blob_get_ranges__when_timeseries_is_empty(table, intervals):
    results = table.blob_get_ranges(_blob_col_name(table), intervals)
    assert len(results) == 2
    assert len(results[0]) == 0
    assert len(results[1]) == 0


def test_blob_erase_ranges__when_timeseries_is_empty(table, intervals):
    erased_count = table.erase_ranges(_blob_col_name(table), intervals)
    assert erased_count == 0


def test_blob_get_ranges(table, intervals):
    inserted_blob_data = _generate_blob_ts(_start_time(intervals), 1000)
    table.blob_insert(_blob_col_name(table),
                      inserted_blob_data[0],
                      inserted_blob_data[1])

    results = table.blob_get_ranges(_blob_col_name(table), [(
        _start_time(intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    _check_ts_results(results, inserted_blob_data, 10)

    results = table.blob_get_ranges(_blob_col_name(table),
                                    [(_start_time(intervals),
                                      _start_time(intervals) + np.timedelta64(10, 's')),
                                     (_start_time(intervals) + np.timedelta64(10, 's'),
                                      _start_time(intervals) + np.timedelta64(20, 's'))])

    _check_ts_results(results, inserted_blob_data, 20)

    # Everything
    results = table.blob_get_ranges(_blob_col_name(table))
    _check_ts_results(results, inserted_blob_data, 1000)

    # empty result
    out_of_time = _start_time(intervals) + np.timedelta64(10, 'h')
    results = table.blob_get_ranges(_blob_col_name(
        table), [(out_of_time, out_of_time + np.timedelta64(10, 's'))])
    assert len(results) == 2
    assert len(results[0]) == 0
    assert len(results[1]) == 0

    # error: column doesn't exist
    with pytest.raises(quasardb.Error):
        table.blob_get_ranges("lolilol", [(_start_time(
            intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    with pytest.raises(quasardb.Error):
        table.blob_insert(
            "lolilol",
            inserted_blob_data[0],
            inserted_blob_data[1])

    with pytest.raises(TypeError):
        table.int64_get_ranges(_blob_col_name, [(_start_time(
            intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    with pytest.raises(TypeError):
        table.int64_insert(
            _blob_col_name(table),
            inserted_blob_data[0],
            inserted_blob_data[1])


def test_blob_erase_ranges(table, intervals):
    inserted_blob_data = _generate_blob_ts(_start_time(intervals), 1000)
    table.blob_insert(
        _blob_col_name(table),
        inserted_blob_data[0],
        inserted_blob_data[1])

    results = table.blob_get_ranges(_blob_col_name(table), [(
        _start_time(intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    erased_count = table.erase_ranges(_blob_col_name(table), [(
        _start_time(intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    assert erased_count == len(results[0])

    erased_count = table.erase_ranges(_blob_col_name(table), [(
        _start_time(intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    assert erased_count == 0

    results = table.blob_get_ranges(_blob_col_name(table), [(
        _start_time(intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    assert len(results[0]) == 0

###
# Blob tests
#


def test_string_get_ranges__when_timeseries_is_empty(table, intervals):
    results = table.string_get_ranges(_blob_col_name(table), intervals)
    assert len(results) == 2
    assert len(results[0]) == 0
    assert len(results[1]) == 0


def test_string_erase_ranges__when_timeseries_is_empty(table, intervals):
    erased_count = table.erase_ranges(_string_col_name(table), intervals)
    assert erased_count == 0


def test_string_get_ranges(table, intervals):
    inserted_string_data = _generate_string_ts(_start_time(intervals), 25)
    table.string_insert(_string_col_name(table),
                        inserted_string_data[0],
                        inserted_string_data[1])

    results = table.string_get_ranges(_string_col_name(table), [(
        _start_time(intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    _check_ts_results(results, inserted_string_data, 10)

    results = table.string_get_ranges(
        _string_col_name(table),
        [
            (_start_time(intervals),
             _start_time(intervals) +
             np.timedelta64(
                10,
                's')),
            (_start_time(intervals) +
             np.timedelta64(
                10,
                's'),
                _start_time(intervals) +
                np.timedelta64(
                20,
                's'))])

    _check_ts_results(results, inserted_string_data, 20)

    # Everything
    results = table.string_get_ranges(_string_col_name(table))
    _check_ts_results(results, inserted_string_data, 25)

    # empty result
    out_of_time = _start_time(intervals) + np.timedelta64(10, 'h')
    results = table.string_get_ranges(_string_col_name(
        table), [(out_of_time, out_of_time + np.timedelta64(10, 's'))])
    assert len(results) == 2
    assert len(results[0]) == 0
    assert len(results[1]) == 0

    # error: column doesn't exist
    with pytest.raises(quasardb.Error):
        table.string_get_ranges("lolilol", [(_start_time(
            intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    with pytest.raises(quasardb.Error):
        table.string_insert(
            "lolilol",
            inserted_string_data[0],
            inserted_string_data[1])

    with pytest.raises(TypeError):
        table.int64_get_ranges(_string_col_name, [(_start_time(
            intervals), _start_time(intervals) + np.timedelta64(10, 's'))])

    with pytest.raises(TypeError):
        table.int64_insert(
            _string_col_name(table),
            inserted_string_data[0],
            inserted_string_data[1])



from time import sleep
import re
import pytest
import quasardb
import quasardb.stats as qdbst
import pprint

import test_batch_inserter as batchlib
import conftest

_has_stats = False


def _write_data(conn, table):
    inserter = conn.inserter(
        batchlib._make_inserter_info(table))

    doubles, blobs, strings, integers, timestamps = batchlib._test_with_table(
        inserter,
        table,
        conftest.create_many_intervals(),
        batchlib._regular_push)


def _ensure_stats(conn, table):
    global _has_stats

    # This function is merely here to generate some activity on the cluster,
    # so that we're guaranteed to have some statistics
    if _has_stats is False:
        _write_data(conn, table)
        # Statistics refresh interval is 5 seconds.
        sleep(6)
        _has_stats = True


# A fairly conservative set of user-specific stats we always expect after doing our
# _write_data() activity generating.
#
# We can add keys here if we want to better test against regressions, but for now it's
# fairly conservative.
_expected_user_stats = ['requests.total_count',
                        'requests.successes_count',
                        'requests.bytes_in',
                        'requests.bytes_out',
                        'perf.ts.table_insert.deserialization.total_ns',
                        'perf.ts.table_insert.processing.total_ns']

# Same, but cumulative stats.
_expected_cumulative_stats = [
    'requests.total_count',
    'requests.successes_count',
    'requests.bytes_in',
    'requests.bytes_out',
    'perf.ts.buffered_table_insert.deserialization.total_ns',
    'perf.ts.buffered_table_insert.processing.total_ns',
    'partitions_count']


def test_stats_by_node(qdbd_secure_connection,
                       secure_table):
    assert(len(_expected_cumulative_stats) > len(_expected_user_stats))
    _ensure_stats(qdbd_secure_connection, secure_table)
    xs = qdbst.by_node(qdbd_secure_connection)

    pp = pprint.PrettyPrinter(indent=4)
    # pp.pprint(xs['127.0.0.1:2838']['cumulative'])

    assert len(xs) == 1

    for _, stats in xs.items():
        assert 'by_uid' in stats
        assert 'cumulative' in stats

        # Test user stats
        for uid, xs in stats['by_uid'].items():
            assert isinstance(uid, int)

            for expected in _expected_user_stats:
                assert expected in xs

            for k, v in xs.items():
                # As far as I know, per-user statistics should *always* be
                # integers
                assert isinstance(v, int)

        # Test cumulative stats
        xs = stats['cumulative']
        for expected in _expected_cumulative_stats:
            assert expected in xs

def test_stats_regex():
    # This is mostly to test a regression, where user-stats for users with multi-digit ids
    # were not picked up.
    user_stats = ['$qdb.statistics.foo.uid_1',
                  '$qdb.statistics.foo.uid_21',
                  '$qdb.statistics.foo.uid_321',
                  '$qdb.statistics.foo.bar.uid_321']
    total_stats = ['$qdb.statistics.foo',
                   '$qdb.statistics.foo.bar',
                   '$qdb.statistics.foo.bar1'
                   '$qdb.statistics.u.i.d.1'
                   '$qdb.statistics.foo.uid1']

    for s in user_stats:
        assert qdbst.is_user_stat(s) is True
        assert qdbst.is_cumulative_stat(s) is False

    for s in total_stats:
        assert qdbst.is_user_stat(s) is False
        assert qdbst.is_cumulative_stat(s) is True



import os
import pdoc
import quasardb
import quasardb.pool
import quasardb.stats
import quasardb.pandas as qdbpd

context = pdoc.Context()

# This is a hack: pydoc has a lot of issues with importing submodules properly. It's
# related to pybind11 generating invalid docstrings, and we would get import errors
# and whatnot. E.g. the quasardb.Blob() has a `expiry=datetime.datetime()` mention
# in the type, _but_ the actual modules (as python sees it) it generates don't even
# import datetime.
#
# This then causes pydoc to throw an error.
#
# As such, we're making our own module subclass here, which will allow us to manually
# tell pybind which submodules to load, rather than telling it to traverse everything
# automatically.
class Module(pdoc.Module):
    def __init__(self, *args, submodules=[], **kwargs):
        super().__init__(*args, **kwargs)
        self._submodules = submodules

    def submodules(self):
        return self._submodules


module_qdb = Module(quasardb.quasardb, context=context,
                    submodules=[pdoc.Module(quasardb.pool, context=context),
                                pdoc.Module(quasardb.stats, context=context),
                                pdoc.Module(quasardb.pandas, context=context)])

modules = [module_qdb]

pdoc.link_inheritance(context)

def recursive_htmls(mod):
    yield mod.name, mod.html()
    for submod in mod.submodules():
        yield from recursive_htmls(submod)

def _strip_prefix(s, p):
    if s.startswith(p):
        return s[len(p):]
    else:
        return s

def write_module(filename, html):
    with open(filename, 'w') as f:
        f.write(html)


for mod in modules:
    for module_name, html in recursive_htmls(mod):
        # This hack is related to the fact that _sometimes_, a module is called
        # `quasardb.quasardb`, and other times it's just `quasardb`. Apparently, when
        # a module's name is `quasardb.pool`, pydoc thinks the file should be called
        # `pool` :/
        #
        # So, just to make sure we have "everything", we write each module twice:
        # once with the quasardb. prefix, and once without.
        write_module("doc/" + module_name + ".html", html)
        write_module("doc/" + _strip_prefix(module_name, 'quasardb.') + ".html", html)

        try:
            os.mkdir("doc/" + _strip_prefix(module_name, 'quasardb.'))
        except FileExistsError:
            pass

        write_module("doc/" + _strip_prefix(module_name, 'quasardb.') + "/index.html", html)



# # pylint: disable=C0103,C0111,C0302,W0212
import pytest
import quasardb
import numpy as np

import test_table as tslib


def _insert_double_points(table, start_time, points=10):
    inserted_double_data = tslib._generate_double_ts(start_time, points)
    table.double_insert(tslib._double_col_name(table),
                        inserted_double_data[0],
                        inserted_double_data[1])
    return inserted_double_data


def _insert_blob_points(table, start_time, points=10):
    inserted_blob_data = tslib._generate_blob_ts(start_time, points)
    table.blob_insert(tslib._blob_col_name(table),
                      inserted_blob_data[0],
                      inserted_blob_data[1])
    return inserted_blob_data


def _insert_int64_points(table, start_time, points=10):
    inserted_int64_data = tslib._generate_int64_ts(start_time, points)
    table.int64_insert(tslib._int64_col_name(table),
                       inserted_int64_data[0],
                       inserted_int64_data[1])
    return inserted_int64_data


def _insert_timestamp_points(table, start_time, points=10):
    inserted_timestamp_data = tslib._generate_timestamp_ts(
        start_time, start_time, points)
    table.timestamp_insert(tslib._ts_col_name(table),
                           inserted_timestamp_data[0],
                           inserted_timestamp_data[1])
    return inserted_timestamp_data

##
# Query failure tests


def test_returns_invalid_argument_for_null_query(qdbd_connection):
    with pytest.raises(TypeError):
        qdbd_connection.query()


def test_returns_invalid_argument_for_empty_query(qdbd_connection):
    with pytest.raises(quasardb.Error):
        qdbd_connection.query('')


def test_returns_invalid_argument_with_invalid_query(qdbd_connection):
    with pytest.raises(quasardb.Error):
        qdbd_connection.query('select * from')


def test_returns_alias_not_found_when_ts_doesnt_exist(qdbd_connection):
    with pytest.raises(quasardb.Error):
        qdbd_connection.query(
            'select * from ' +
            'this_ts_doesnt_exist' +
            ' in range(2017, +10d)')


def test_returns_alias_not_found_when_untagged(qdbd_connection, tag_name):
    with pytest.raises(quasardb.Error):
        qdbd_connection.query(
            "select * from find(tag='" + tag_name + "') in range(2017, +10d)")


def test_returns_columns_not_found(qdbd_connection, table, column_name):
    with pytest.raises(quasardb.Error):
        qdbd_connection.query(
            "select " +
            column_name +
            " from " +
            table.get_name() +
            " in range(2017, +10d)")

def test_invalid_query_message(qdbd_connection, table):
    with pytest.raises(quasardb.Error) as excinfo:
        qdbd_connection.query(
            "select asd_col from \"" +
            table.get_name() +
            "\" in range(2016-01-01 , 2016-12-12)")
    assert str(excinfo.value) == 'The timeseries does not contain this column. Could not find column \'asd_col\'.'


##
# Double data tests

def sanity_check(ts_name, scanned_point_count, res, rows_count, columns_count):
    assert res.scanned_point_count == scanned_point_count
    assert len(res.tables) == 1
    assert len(res.tables[ts_name][0].data) == rows_count
    assert len(res.tables[ts_name]) == columns_count
    assert res.tables[ts_name][0].name == "$timestamp"


def test_returns_empty_result(qdbd_connection, table):
    res = qdbd_connection.query(
        "select * from \"" +
        table.get_name() +
        "\" in range(2016-01-01 , 2016-12-12)")
    assert len(res) == 0


def test_returns_table_as_string(
        qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    query = "select * from \"" + table.get_name() + "\" in range(" + \
        str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query)

    assert len(res) == 10

    for row, _ in zip(res, inserted_double_data[1]):
        assert row['$table'] == table.get_name()


def test_returns_table_as_blob(
        qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    query = "select * from \"" + table.get_name() + "\" in range(" + \
        str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query, blobs=['$table'])

    assert len(res) == 10

    for row, _ in zip(res, inserted_double_data[1]):
        assert row['$table'] == table.get_name()


def test_returns_inserted_data_with_star_select(
        qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    query = "select * from \"" + table.get_name() + "\" in range(" + \
        str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query)

    assert len(res) == 10

    for row, v in zip(res, inserted_double_data[1]):
        assert '$timestamp' in row

        assert row['the_blob'] is None
        assert row['the_int64'] is None
        assert row['the_ts'] is None
        assert row['$table'] == table.get_name()
        assert row['the_double'] == v


def test_returns_inserted_data_with_column_select(
        qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    query = "select " + tslib._double_col_name(table) + \
        " from \"" + table.get_name() + "\" in range(" + \
        str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query)

    assert len(res) == 10

    for row, v in zip(res, inserted_double_data[1]):
        assert '$table' not in row
        assert '$timestamp' not in row
        assert 'the_blob' not in row
        assert 'the_int64' not in row
        assert 'the_ts' not in row

        assert row['the_double'] == v


def test_returns_inserted_data_with_specific_select(
        qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    query = "select $timestamp, $table, " + tslib._double_col_name(table) + \
        " from \"" + table.get_name() + "\" in range(" + \
        str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query)

    assert len(res) == 10

    for row, v in zip(res, inserted_double_data[1]):
        assert '$timestamp' in row
        assert 'the_blob' not in row
        assert 'the_int64' not in row
        assert 'the_ts' not in row

        assert row['the_double'] == v


def test_returns_count_data_with_count_select(
        qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    _ = _insert_double_points(table, start_time, 10)
    query = "select count(" + tslib._double_col_name(table) + ") from \"" + table.get_name() + \
        "\" in range(" + str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query)

    assert len(res) == 1
    assert res[0]['count(the_double)'] == 10


def test_returns_count_data_with_sum_select(
        qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 10)
    query = "select sum(" + tslib._double_col_name(table) + ") from \"" + table.get_name() + \
        "\" in range(" + str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query)

    assert len(res) == 1
    assert pytest.approx(res[0]['sum(the_double)'],
                         np.sum(inserted_double_data[1]))


def test_returns_inserted_multi_data_with_star_select(
        qdbd_connection, table, intervals):
    start_time = tslib._start_time(intervals)
    inserted_double_data = _insert_double_points(table, start_time, 100)
    inserted_blob_data = _insert_blob_points(table, start_time, 100)
    inserted_int64_data = _insert_int64_points(table, start_time, 100)
    inserted_timestamp_data = _insert_timestamp_points(table, start_time, 100)

    query = "select * from \"" + table.get_name() + \
        "\" in range(" + str(tslib._start_year(intervals)) + ", +100d)"
    res = qdbd_connection.query(query, blobs=['the_blob'])

    assert len(res) == 100

    print("blobs: ", type(inserted_blob_data[1][0]))

    for row, double, blob, int64, ts in zip(res,
                                            inserted_double_data[1],
                                            inserted_blob_data[1],
                                            inserted_int64_data[1],
                                            inserted_timestamp_data[1]):
        assert '$timestamp' in row

        # Note that this is a string
        assert row['$table'] == table.get_name()

        # And this is a blob
        assert row['the_blob'] == blob

        assert row['the_double'] == double
        assert row['the_int64'] == int64
        assert row['the_ts'] == ts


def test_create_table(qdbd_connection, entry_name):
    query = "create table \"{}\" (col int64)".format(entry_name)
    _ = qdbd_connection.query(query)



# Copyright (c) 2009-2021, quasardb SAS. All rights reserved.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of quasardb nor the names of its contributors may
#      be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY QUASARDB AND CONTRIBUTORS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from __future__ import print_function
from builtins import int

import os
import sys
import traceback
import datetime

import quasardb  # pylint: disable=C0413,E0401
import numpy as np

def main(quasardb_uri, blob_name):
    print("Connecting to: ", quasardb_uri)
    c = quasardb.Cluster(uri=quasardb_uri)

    # create a blob object
    b = c.blob(blob_name)

    # add data to the blob (which may, or may not exist)
    b.update('content')

    # print the content
    print("content:", b.get())

    # you can set expiry to blobs
    # by default there is none
    print("expiry =", b.get_expiry_time())

    # absolute expiry
    b.expires_at(datetime.datetime.now() + datetime.timedelta(minutes=2))
    print("expiry =", b.get_expiry_time())

    # relative expiry
    b.expires_from_now(datetime.timedelta(minutes=2))
    print("expiry =", b.get_expiry_time())

    b.remove()

if __name__ == "__main__":

    try:
        if len(sys.argv) != 3:
            print("usage: ", sys.argv[0], " quasardb_uri blob_name")
            sys.exit(1)

        main(sys.argv[1], sys.argv[2])

    except Exception as ex:  # pylint: disable=W0703
        print("An error ocurred:", str(ex))
        traceback.print_exc()



# pylint: disable=C0103,C0111,C0302,W0212
from builtins import range as xrange, int as long  # pylint: disable=W0622
from functools import reduce  # pylint: disable=W0622
import datetime
import test_table as tslib
from time import sleep

import pytest
import quasardb
import numpy as np


def _row_insertion_method(
        inserter,
        dates,
        doubles,
        blobs,
        strings,
        integers,
        timestamps):
    for i in range(len(dates)):
        inserter.start_row(dates[i])
        inserter.set_double(0, doubles[i])
        inserter.set_blob(1, blobs[i])
        inserter.set_string(2, strings[i])
        inserter.set_int64(3, integers[i])
        inserter.set_timestamp(4, timestamps[i])


def _regular_push(inserter):
    inserter.push()


def _async_push(inserter):
    inserter.push_async()
    # Wait for push_async to complete
    # Ideally we could be able to get the proper flush interval
    sleep(15)


def _fast_push(inserter):
    inserter.push_fast()


def _make_inserter_info(table):
    return [quasardb.BatchColumnInfo(table.get_name(), tslib._double_col_name(table), 100),
            quasardb.BatchColumnInfo(
                table.get_name(), tslib._blob_col_name(table), 100),
            quasardb.BatchColumnInfo(
                table.get_name(), tslib._string_col_name(table), 100),
            quasardb.BatchColumnInfo(
                table.get_name(), tslib._int64_col_name(table), 100),
            quasardb.BatchColumnInfo(table.get_name(), tslib._ts_col_name(table), 100)]


def test_non_existing_bulk_insert(qdbd_connection, entry_name):
    with pytest.raises(quasardb.Error):
        qdbd_connection.inserter(
            [quasardb.BatchColumnInfo(entry_name, "col", 10)])


def _generate_data(count, start=np.datetime64('2017-01-01', 'ns')):
    doubles = np.random.uniform(-100.0, 100.0, count)
    integers = np.random.randint(-100, 100, count)
    blobs = np.array(
        list(
            np.random.bytes(
                np.random.randint(
                    8,
                    16)) for i in range(count)),
        'O')
    strings = np.array([("content_" + str(item)) for item in range(count)])
    timestamps = tslib._generate_dates(
        start + np.timedelta64('1', 'D'), count)

    return (doubles, integers, blobs, strings, timestamps)


def _set_batch_inserter_data(inserter, intervals, data, start=0):
    (doubles, integers, blobs, strings, timestamps) = data

    for i in range(start, len(intervals)):
        inserter.start_row(intervals[i])
        inserter.set_double(0, doubles[i])
        inserter.set_blob(1, blobs[i])
        inserter.set_string(2, strings[i])
        inserter.set_int64(3, integers[i])
        inserter.set_timestamp(4, timestamps[i])


def _assert_results(table, intervals, data):
    (doubles, integers, blobs, strings, timestamps) = data

    whole_range = (intervals[0], intervals[-1:][0] + np.timedelta64(2, 's'))
    results = table.double_get_ranges(
        tslib._double_col_name(table), [whole_range])

    np.testing.assert_array_equal(results[0], intervals)
    np.testing.assert_array_equal(results[1], doubles)

    results = table.blob_get_ranges(tslib._blob_col_name(table), [whole_range])
    np.testing.assert_array_equal(results[0], intervals)
    np.testing.assert_array_equal(results[1], blobs)

    results = table.string_get_ranges(
        tslib._string_col_name(table), [whole_range])
    np.testing.assert_array_equal(results[0], intervals)
    np.testing.assert_array_equal(results[1], strings)

    results = table.int64_get_ranges(
        tslib._int64_col_name(table), [whole_range])
    np.testing.assert_array_equal(results[0], intervals)
    np.testing.assert_array_equal(results[1], integers)

    results = table.timestamp_get_ranges(
        tslib._ts_col_name(table), [whole_range])
    np.testing.assert_array_equal(results[0], intervals)
    np.testing.assert_array_equal(results[1], timestamps)


def _test_with_table(
        inserter,
        table,
        intervals,
        push_method=_regular_push,
        data=None):

    if data is None:
        data = _generate_data(len(intervals))

    # range is right exclusive, so the timestamp has to be beyond
    whole_range = (intervals[0], intervals[-1:][0] + np.timedelta64(2, 's'))

    (doubles, integers, blobs, strings, timestamps) = data

    _set_batch_inserter_data(inserter, intervals, data)

    # before the push, there is nothing
    results = table.double_get_ranges(
        tslib._double_col_name(table), [whole_range])
    assert len(results[0]) == 0

    results = table.blob_get_ranges(tslib._blob_col_name(table), [whole_range])
    assert len(results[0]) == 0

    results = table.string_get_ranges(
        tslib._string_col_name(table), [whole_range])
    assert len(results[0]) == 0

    results = table.int64_get_ranges(
        tslib._int64_col_name(table), [whole_range])
    assert len(results[0]) == 0

    results = table.timestamp_get_ranges(
        tslib._ts_col_name(table), [whole_range])
    assert len(results[0]) == 0

    # after push, there is everything
    push_method(inserter)
    if push_method == _async_push:
        sleep(20)

    _assert_results(table, intervals, data)

    return doubles, blobs, strings, integers, timestamps


def test_successful_bulk_row_insert(qdbd_connection, table, many_intervals):
    inserter = qdbd_connection.inserter(_make_inserter_info(table))

    _test_with_table(
        inserter,
        table,
        many_intervals,
        _regular_push)


def test_successful_secure_bulk_row_insert(
        qdbd_secure_connection,
        secure_table,
        many_intervals):
    inserter = qdbd_secure_connection.inserter(
        _make_inserter_info(secure_table))

    _test_with_table(
        inserter,
        secure_table,
        many_intervals,
        _regular_push)


def test_successful_async_bulk_row_insert(
        qdbd_connection, table, many_intervals):

    # Same test as `test_successful_bulk_row_insert` but using `push_async` to push the entries
    # This allows us to test the `push_async` feature

    inserter = qdbd_connection.inserter(_make_inserter_info(table))
    _test_with_table(
        inserter,
        table,
        many_intervals,
        _async_push)


def test_successful_fast_bulk_row_insert(
        qdbd_connection, table, many_intervals):
    # Same test as `test_successful_bulk_row_insert` but using `push_async` to push the entries
    # This allows us to test the `push_async` feature

    inserter = qdbd_connection.inserter(_make_inserter_info(table))
    _test_with_table(
        inserter,
        table,
        many_intervals,
        _fast_push)


def test_failed_local_table_with_wrong_columns(qdbd_connection, entry_name):
    columns = [quasardb.BatchColumnInfo(entry_name, "1000flavorsofwrong", 10)]
    with pytest.raises(quasardb.Error):
        qdbd_connection.inserter(columns)


def test_push_truncate_implicit_range(qdbd_connection, table, many_intervals):

    whole_range = (
        many_intervals[0], many_intervals[-1:][0] + np.timedelta64(2, 's'))

    # Generate our dataset
    data = _generate_data(len(many_intervals))
    (doubles, integers, blobs, strings, timestamps) = data

    # Insert once
    inserter = qdbd_connection.inserter(_make_inserter_info(table))
    _set_batch_inserter_data(inserter, many_intervals, data)
    inserter.push()

    # Compare results, should be equal
    results = table.double_get_ranges(
        tslib._double_col_name(table), [whole_range])

    np.testing.assert_array_equal(results[0], many_intervals)
    np.testing.assert_array_equal(results[1], doubles)

    # Insert regular, twice
    _set_batch_inserter_data(inserter, many_intervals, data)
    inserter.push()

    # Compare results, should now have the same data twice
    results = table.double_get_ranges(
        tslib._double_col_name(table), [whole_range])

    assert len(results[1]) == 2 * len(doubles)

    # Insert truncate, should now have original data again
    _set_batch_inserter_data(inserter, many_intervals, data)
    inserter.push_truncate()

    # Verify results, truncating should now make things the same
    # as the beginning again.
    results = table.double_get_ranges(
        tslib._double_col_name(table), [whole_range])

    np.testing.assert_array_equal(results[0], many_intervals)
    np.testing.assert_array_equal(results[1], doubles)


def test_push_truncate_explicit_range(qdbd_connection, table, many_intervals):

    whole_range = (
        many_intervals[0], many_intervals[-1:][0] + np.timedelta64(2, 's'))

    # Generate our dataset
    data = _generate_data(len(many_intervals))
    (doubles, integers, blobs, strings, timestamps) = data

    inserter = qdbd_connection.inserter(_make_inserter_info(table))

    # Insert once
    truncate_range = (whole_range[0],
                      whole_range[1] + np.timedelta64(1, 'ns'))

    _set_batch_inserter_data(inserter, many_intervals, data)
    inserter.push()

    # Verify results, truncating should now make things the same
    # as the beginning again.
    results = table.double_get_ranges(
        tslib._double_col_name(table), [whole_range])

    np.testing.assert_array_equal(results[0], many_intervals)
    np.testing.assert_array_equal(results[1], doubles)

    # If we now set the same data, skip the first element, but keep
    # the same time range, the first element will *not* be present in
    # the resulting dataset.
    _set_batch_inserter_data(inserter, many_intervals, data, start=1)
    inserter.push_truncate(range=truncate_range)

    # Verify results, truncating should now make things the same
    # as the beginning again.
    results = table.double_get_ranges(
        tslib._double_col_name(table), [whole_range])

    np.testing.assert_array_equal(results[0], many_intervals[1:])
    np.testing.assert_array_equal(results[1], doubles[1:])


def test_push_truncate_throws_error_on_invalid_range(
        qdbd_connection, table, many_intervals):
    whole_range = (
        many_intervals[0], many_intervals[-1:][0] + np.timedelta64(2, 's'))

    # Generate our dataset
    data = _generate_data(len(many_intervals))
    (doubles, integers, blobs, strings, timestamps) = data

    # Insert truncate with explicit timerange, we point the start right after the
    # first element in our dataset. This means that the range does not overlap all
    # the data anymore.
    truncate_range = (whole_range[0] + np.timedelta64(1, 'ns'),
                      whole_range[1] + np.timedelta64(1, 'ns'))

    inserter = qdbd_connection.inserter(_make_inserter_info(table))
    _set_batch_inserter_data(inserter, many_intervals, data)
    with pytest.raises(quasardb.Error):
        inserter.push_truncate(range=truncate_range)



# pylint: disable=C0103,C0111,C0302,W0212
import pytest
import quasardb
from functools import reduce
import test_batch_inserter as batchlib
import numpy as np


def test_reader_can_return_no_rows(qdbd_connection, table, many_intervals):
    assert 0 == reduce(lambda x, y: x + 1, table.reader(), 0)


def test_reader_returns_correct_results(
        qdbd_connection, table, many_intervals):
    inserter = qdbd_connection.inserter(
        batchlib._make_inserter_info(table))

    doubles, blobs, strings, integers, timestamps = batchlib._test_with_table(
        inserter,
        table,
        many_intervals,
        batchlib._regular_push)

    offset = 0
    for row in table.reader():
        assert row[1] == doubles[offset]
        assert row[2] == blobs[offset]
        assert row[3] == strings[offset]
        assert row[4] == integers[offset]
        assert row[5] == timestamps[offset]

        offset = offset + 1


def test_reader_iterator_returns_reference(
        qdbd_connection, table, many_intervals, capsys):
    # For performance reasons, our iterators are merely references to the
    # underlying local_table position. A side effect is that if the iterator moves
    # forward, all references will move forward as well.
    #
    # This is actually undesired, and this test is here to detect regressions in
    # behavior.
    inserter = qdbd_connection.inserter(
        batchlib._make_inserter_info(table))

    doubles, blobs, strings, integers, timestamps = batchlib._test_with_table(
        inserter,
        table,
        many_intervals,
        batchlib._regular_push)

    rows = []
    for row in table.reader():
        rows.append(row)

    for row in rows:
        # Timestamp is copied by value
        assert isinstance(row[0], np.datetime64)

        assert row[1] is None
        assert row[2] is None
        assert row[3] is None
        assert row[4] is None
        assert row[5] is None


def test_reader_can_copy_rows(qdbd_connection, table, many_intervals):
    # As a mitigation to the local table reference issue tested above,
    # we provide the ability to copy rows.

    inserter = qdbd_connection.inserter(
        batchlib._make_inserter_info(table))

    doubles, blobs, strings, integers, timestamps = batchlib._test_with_table(
        inserter,
        table,
        many_intervals,
        batchlib._regular_push)

    rows = []
    for row in table.reader():
        copied = row.copy()
        rows.append(copied)

    offset = 0
    for row in rows:
        assert row[1] == doubles[offset]
        assert row[2] == blobs[offset]
        assert row[3] == strings[offset]
        assert row[4] == integers[offset]
        assert row[5] == timestamps[offset]

        offset = offset + 1


def test_reader_can_select_columns(qdbd_connection, table, many_intervals):
    # Verifies that we can select a subset of the total available columns.
    inserter = qdbd_connection.inserter(
        batchlib._make_inserter_info(table))

    doubles, blobs, strings, integers, timestamps = batchlib._test_with_table(
        inserter,
        table,
        many_intervals,
        batchlib._regular_push)

    offset = 0
    for row in table.reader(columns=['the_int64', 'the_double']):
        assert row[1] == integers[offset]
        assert row[2] == doubles[offset]

        with pytest.raises(IndexError):
            print(str(row[3]))

        offset = offset + 1


def test_reader_can_request_ranges(qdbd_connection, table, many_intervals):
    # Verifies that we can select ranges
    inserter = qdbd_connection.inserter(
        batchlib._make_inserter_info(table))

    doubles, blobs, string, integers, timestamps = batchlib._test_with_table(
        inserter,
        table,
        many_intervals,
        batchlib._regular_push)

    first_range = (many_intervals[0], many_intervals[1])
    second_range = (many_intervals[1], many_intervals[2])
    assert 1 == reduce(
        lambda x,
        y: x + 1,
        table.reader(
            ranges=[first_range]),
        0)
    assert 2 == reduce(
        lambda x,
        y: x + 1,
        table.reader(
            ranges=[
                first_range,
                second_range]),
        0)


def test_reader_raises_error_on_invalid_datetime_ranges(
        qdbd_connection, table, many_intervals):
    # Verifies that we can select ranges
    inserter = qdbd_connection.inserter(
        batchlib._make_inserter_info(table))

    doubles, blobs, strings, integers, timestamps = batchlib._test_with_table(
        inserter,
        table,
        many_intervals,
        batchlib._regular_push)

    ns1 = many_intervals[0]
    ns2 = many_intervals[1]

    s1 = np.datetime64(ns1, 's')
    s2 = np.datetime64(ns2, 's')

    # Works
    r = table.reader(ranges=[(ns1, ns2)])

    with pytest.raises(quasardb.quasardb.InvalidDatetimeError):
        r = table.reader(ranges=[(s1, ns2)])

    with pytest.raises(quasardb.quasardb.InvalidDatetimeError):
        r = table.reader(ranges=[(ns1, s2)])

    with pytest.raises(quasardb.quasardb.InvalidDatetimeError):
        r = table.reader(ranges=[(s1, s2)])


def test_reader_can_read_dicts(qdbd_connection, table, many_intervals):
    # Verifies that we can select a subset of the total available columns.
    inserter = qdbd_connection.inserter(
        batchlib._make_inserter_info(table))

    doubles, blobs, strings, integers, timestamps = batchlib._test_with_table(
        inserter,
        table,
        many_intervals,
        batchlib._regular_push)

    offset = 0
    for row in table.reader(dict=True):
        assert isinstance(row['$timestamp'], np.datetime64)
        assert row['the_double'] == doubles[offset]
        assert row['the_blob'] == blobs[offset]
        assert row['the_string'] == strings[offset]
        assert row['the_int64'] == integers[offset]
        assert row['the_ts'] == timestamps[offset]

        with pytest.raises(KeyError):
            print(str(row['nothere']))

        offset = offset + 1


def test_reader_can_copy_dict_rows(qdbd_connection, table, many_intervals):
    # Just like our regular row reader, our dict-based ready also needs to
    # copy the rows.
    inserter = qdbd_connection.inserter(
        batchlib._make_inserter_info(table))

    doubles, blobs, strings, integers, timestamps = batchlib._test_with_table(
        inserter,
        table,
        many_intervals,
        batchlib._regular_push)

    offset = 0
    rows = []

    for row in table.reader(dict=True):
        copied = row.copy()
        rows.append(copied)

    for row in rows:
        assert row['the_double'] == doubles[offset]
        assert row['the_blob'] == blobs[offset]
        assert row['the_string'] == strings[offset]
        assert row['the_int64'] == integers[offset]
        assert row['the_ts'] == timestamps[offset]

        with pytest.raises(KeyError):
            print(str(row['nothere']))

        offset = offset + 1



# pylint: disable=C0103,C0111,C0302,W0212
import pytest
import quasardb


def test_types(qdbd_connection, blob_entry, random_blob, tag_name):

    blob_entry.put(random_blob)
    blob_entry.attach_tag(tag_name)

    res = qdbd_connection.find("find(tag='" + tag_name + "')").run()

    assert len(res) == 1

    res = qdbd_connection.find(
        "find(tag='" + tag_name + "' AND type=blob)").run()

    assert len(res) == 1

    res = qdbd_connection.find(
        "find(tag='" + tag_name + "' AND type=integer)").run()

    assert len(res) == 0


def test_types(qdbd_connection, blob_entry, random_blob, tag_names):
    blob_entry.put(random_blob)
    blob_entry.attach_tags(tag_names)

    res = qdbd_connection.find("find(tag='" + tag_names[0] + "')").run()
    assert len(res) == 1

    res = qdbd_connection.find("find(tag='" + tag_names[1] + "')").run()
    assert len(res) == 1

    res = qdbd_connection.find(
        "find(tag='" +
        tag_names[0] +
        "' AND tag='" +
        tag_names[1] +
        "')").run()

    assert len(res) == 1



# pylint: disable=missing-module-docstring,missing-function-docstring,not-an-iterable,invalid-name

from datetime import timedelta
import logging
import numpy as np
import numpy.ma as ma
import pandas as pd
import pytest

import quasardb
import quasardb.pandas as qdbpd

ROW_COUNT = 1000


def gen_df(start_time, count, step=1):
    idx = pd.date_range(start=start_time,
                        end=(start_time + (np.timedelta64(step, 'D') * count)),
                        periods=count)

    return pd.DataFrame(data={"the_double": np.random.uniform(-100.0, 100.0, count),
                              "the_int64": np.random.randint(-100, 100, count),
                              "the_blob": np.array(list(np.random.bytes(np.random.randint(16, 32)) for i in range(count)),
                                                   'O'),
                              "the_string": np.array([("content_" + str(item))
                                                      for item in range(count)], 'U'),
                              "the_ts": np.array([
                                  (start_time + np.timedelta64(i, 's'))
                                  for i in range(count)]).astype('datetime64[ns]')},
                        index=idx)


def gen_series(start_time, count):
    idx = pd.date_range(start_time, periods=count, freq='S')

    return {"the_double": pd.Series(np.random.uniform(-100.0, 100.0, count),
                                    index=idx),
            "the_int64": pd.Series(np.random.randint(-100, 100, count),
                                   index=idx),
            "the_blob": pd.Series(np.array(list(np.random.bytes(np.random.randint(16, 32)) for i in range(count)),
                                           'O'),
                                  index=idx),
            "the_string": pd.Series(np.array([("content_" + str(item)) for item in range(count)],
                                             'U'),
                                    index=idx),
            "the_ts": pd.Series(np.array([(start_time + np.timedelta64(i, 's'))
                                          for i in range(count)]).astype('datetime64[ns]'),
                                index=idx)}


def test_series_read_write(table):
    series = gen_series(np.datetime64('2017-01-01', 'ns'), ROW_COUNT)

    # Insert everything
    for col in series:
        qdbpd.write_series(series[col], table, col)

    # Read everything
    for col in series:
        read_series = qdbpd.read_series(table, col)
        assert isinstance(read_series, pd.core.series.Series)
        assert read_series.size == ROW_COUNT
        np.testing.assert_array_equal(
            read_series.to_numpy(), series[col].to_numpy())

@pytest.mark.parametrize("write_fn", [qdbpd.write_dataframe,
                                      qdbpd.write_pinned_dataframe])
def test_dataframe(write_fn, qdbd_connection, table):
    df1 = gen_df(np.datetime64('2017-01-01'), ROW_COUNT)
    write_fn(df1, qdbd_connection, table)

    df2 = qdbpd.read_dataframe(table)

    assert len(df1.columns) == len(df2.columns)
    for col in df1.columns:
        np.testing.assert_array_equal(df1[col].to_numpy(), df2[col].to_numpy())


@pytest.mark.parametrize("write_fn", [qdbpd.write_dataframe,
                                      qdbpd.write_pinned_dataframe])
def test_dataframe_can_read_columns(write_fn, qdbd_connection, table):
    df1 = gen_df(np.datetime64('2017-01-01'), ROW_COUNT)
    write_fn(df1, qdbd_connection, table)

    df2 = qdbpd.read_dataframe(table, columns=['the_double', 'the_int64'])

    assert len(df1.columns) != len(df2.columns)
    assert len(df2.columns) == 2
    for col in df2.columns:
        np.testing.assert_array_equal(df1[col].to_numpy(), df2[col].to_numpy())


@pytest.mark.parametrize("write_fn", [qdbpd.write_dataframe,
                                      qdbpd.write_pinned_dataframe])
def test_dataframe_can_read_ranges(write_fn, qdbd_connection, table):
    start = np.datetime64('2017-01-01', 'ns')
    df1 = gen_df(start, 10)
    write_fn(df1, qdbd_connection, table)

    first_range = (start, start + np.timedelta64(1, 'D'))
    second_range = (start + np.timedelta64(1, 'D'),
                    start + np.timedelta64(2, 'D'))

    df2 = qdbpd.read_dataframe(table)
    df3 = qdbpd.read_dataframe(table, ranges=[first_range])
    df4 = qdbpd.read_dataframe(table, ranges=[first_range, second_range])

    assert df2.shape[0] == 10
    assert df3.shape[0] == 1
    assert df4.shape[0] == 2


@pytest.mark.parametrize("write_fn", [qdbpd.write_dataframe,
                                      qdbpd.write_pinned_dataframe])
def test_write_dataframe(write_fn, qdbd_connection, table):
    # Ensures that we can do a full-circle write and read of a dataframe
    df1 = gen_df(np.datetime64('2017-01-01'), ROW_COUNT)
    write_fn(df1, qdbd_connection, table)

    df2 = qdbpd.read_dataframe(table)

    assert len(df1.columns) == len(df2.columns)
    for col in df1.columns:
        np.testing.assert_array_equal(df1[col].to_numpy(), df2[col].to_numpy())

@pytest.mark.parametrize("write_fn", [qdbpd.write_dataframe,
                                      qdbpd.write_pinned_dataframe])
@pytest.mark.parametrize("row_count", [1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024])
@pytest.mark.parametrize("df_count", [1, 2, 4, 8, 16])
def test_write_unindexed_dataframe(write_fn, df_count, row_count, qdbd_connection, table):
    # CF. QDB-10203, we generate a dataframe which is unordered. The easiest
    # way to do this is reuse our gen_df function with overlapping timestamps,
    # and concatenate the data.

    # We generate a whole bunch of dataframes. Their timestamps overlap, but never collide: we
    # ensure that by using a 'step' of the dataframe count, and let each dataframe start at
    # the appropriate offset.
    #
    # e.g. with 8 dataframes, the first dataframe starts at 2017-01-01 and the second row's
    # timestap is 2017-01-09, while the second dataframe starts at 2017-01-02 and has 2017-01-10
    # as second row, etc.
    dfs = [gen_df((np.datetime64('2017-01-01') + np.timedelta64(i, 'D')) ,
                  row_count,
                  step=df_count)
           for i in range(df_count)]
    df_unsorted = pd.concat(dfs)
    df_sorted = df_unsorted.sort_index().reindex()
    assert len(df_sorted.columns) == len(df_unsorted.columns)

    # We store unsorted
    write_fn(df_unsorted, qdbd_connection, table)


    # We expect to receive sorted data
    df_read = qdbpd.read_dataframe(table)

    assert len(df_sorted.columns) == len(df_read.columns)
    for col in df_sorted.columns:
        np.testing.assert_array_equal(df_sorted[col].to_numpy(),
                                      df_read[col].to_numpy())


@pytest.mark.parametrize("write_fn", [qdbpd.write_dataframe,
                                      qdbpd.write_pinned_dataframe])
def test_write_dataframe_push_fast(write_fn, qdbd_connection, table):
    # Ensures that we can do a full-circle write and read of a dataframe
    df1 = gen_df(np.datetime64('2017-01-01'), ROW_COUNT)
    write_fn(df1, qdbd_connection, table, fast=True)

    df2 = qdbpd.read_dataframe(table)

    assert len(df1.columns) == len(df2.columns)
    for col in df1.columns:
        np.testing.assert_array_equal(df1[col].to_numpy(), df2[col].to_numpy())


@pytest.mark.parametrize("write_fn", [qdbpd.write_dataframe,
                                      qdbpd.write_pinned_dataframe])
@pytest.mark.parametrize("truncate", [True,
                                      (np.datetime64('2017-01-01', 'ns'),
                                       np.datetime64('2020-01-01', 'ns'))])
def test_write_dataframe_push_truncate(write_fn, truncate, qdbd_connection, table):
    # Ensures that we can do a full-circle write and read of a dataframe
    df1 = gen_df(np.datetime64('2017-01-01'), ROW_COUNT)

    if type(truncate) == tuple:
        # If this fails, it means our `gen_df` function generated data that goes
        # beyond the (hardcoded) truncate end range.
        #
        # You'll either want to set the `np.datetime64` end truncate time to a later
        # timestamp, or adjust the `gen_df` dataframe.
        assert df1.index[-1] < truncate[1]

    write_fn(df1, qdbd_connection, table, truncate=truncate)
    write_fn(df1, qdbd_connection, table, truncate=truncate)

    df2 = qdbpd.read_dataframe(table)

    assert len(df1.columns) == len(df2.columns)
    for col in df1.columns:
        np.testing.assert_array_equal(df1[col].to_numpy(), df2[col].to_numpy())


@pytest.mark.parametrize("write_fn", [qdbpd.write_dataframe,
                                      qdbpd.write_pinned_dataframe])
def test_write_dataframe_create_table(write_fn, caplog, qdbd_connection, entry_name):
    caplog.set_level(logging.DEBUG)
    table = qdbd_connection.ts(entry_name)
    df1 = gen_df(np.datetime64('2017-01-01'), ROW_COUNT)
    write_fn(df1, qdbd_connection, table, create=True)

    df2 = qdbpd.read_dataframe(table)

    assert len(df1.columns) == len(df2.columns)
    for col in df1.columns:
        np.testing.assert_array_equal(df1[col].to_numpy(), df2[col].to_numpy())


@pytest.mark.parametrize("write_fn", [qdbpd.write_dataframe,
                                      qdbpd.write_pinned_dataframe])
def test_write_dataframe_create_table_twice(write_fn, qdbd_connection, table):
    df1 = gen_df(np.datetime64('2017-01-01'), ROW_COUNT)
    write_fn(df1, qdbd_connection, table, create=True)

@pytest.mark.parametrize("write_fn", [qdbpd.write_dataframe,
                                      qdbpd.write_pinned_dataframe])
def test_write_dataframe_create_table_with_shard_size(write_fn, caplog, qdbd_connection, entry_name):
    caplog.set_level(logging.DEBUG)
    table = qdbd_connection.ts(entry_name)
    df1 = gen_df(np.datetime64('2017-01-01'), ROW_COUNT)
    write_fn(df1, qdbd_connection, table, create=True, shard_size=timedelta(weeks=4))

    df2 = qdbpd.read_dataframe(table)

    assert len(df1.columns) == len(df2.columns)
    for col in df1.columns:
        np.testing.assert_array_equal(df1[col].to_numpy(), df2[col].to_numpy())


def check_equal(expected, actual):
    if np.isnan(expected):
        assert np.isnan(actual)
    else:
        assert expected == actual

@pytest.mark.parametrize("write_fn", [qdbpd.write_dataframe,
                                      qdbpd.write_pinned_dataframe])
def test_dataframe_read_fast_is_unordered(write_fn, qdbd_connection, table):
    # As of now, when reading a dataframe fast, when it contains null values,
    # rows are not guaranteed to be ordered; they might be matched to the
    # wrong rows.
    #
    # This will be fixed when the reader supports pinned columns, so we can
    # read the data fast (as numpy arrays) and at the same time keep ordering.

    df1 = gen_df(np.datetime64('2017-01-01'), 2)
    df2 = gen_df(np.datetime64('2017-01-01'), 2)

    np.testing.assert_array_equal(df1.index, df2.index)

    ts1 = df1.index[0]
    ts2 = df1.index[1]

    # Now, we set the wrong value for a first row in df1, and for
    # a second row in df2
    df1.at[ts1, 'the_double'] = None
    df2.at[ts2, 'the_double'] = None

    df3 = pd.concat([df1, df2]).sort_index()
    write_fn(df3, qdbd_connection, table)

    df4 = qdbpd.read_dataframe(table)

    assert len(df3.columns) == len(df4.columns)
    for col in df3.columns:
        if col == 'the_double':
            np.testing.assert_array_almost_equal(df3[col].to_numpy(),
                                                 df4[col].to_numpy())
        else:
            np.testing.assert_array_equal(df3[col].to_numpy(),
                                          df4[col].to_numpy())

    df5 = qdbpd.read_dataframe(table, row_index=True)

    assert df5.loc[0, 'the_int64'] == df1.loc[ts1, 'the_int64']
    assert df5.loc[1, 'the_int64'] == df2.loc[ts1, 'the_int64']
    assert df5.loc[2, 'the_int64'] == df1.loc[ts2, 'the_int64']
    assert df5.loc[3, 'the_int64'] == df2.loc[ts2, 'the_int64']

    # QDB-2418
    check_equal(df5.loc[0, 'the_double'], df1.loc[ts1, 'the_double'])
    check_equal(df5.loc[1, 'the_double'], df2.loc[ts1, 'the_double'])
    check_equal(df5.loc[2, 'the_double'], df1.loc[ts2, 'the_double'])
    check_equal(df5.loc[3, 'the_double'], df2.loc[ts2, 'the_double'])


@pytest.mark.parametrize("write_fn", [qdbpd.write_dataframe,
                                      qdbpd.write_pinned_dataframe])
def test_query(write_fn, qdbd_connection, table):
    df1 = gen_df(np.datetime64('2017-01-01'), ROW_COUNT)
    write_fn(df1, qdbd_connection, table)

    res = qdbpd.query(qdbd_connection, "SELECT * FROM \"" +
                      table.get_name() + "\"", blobs=['the_blob'])

    for col in df1.columns:
        np.testing.assert_array_equal(df1[col].to_numpy(), res[col].to_numpy())


def _gen_floating(n, low=-100.0, high=100.0):
    return np.random.uniform(low, high, n)


def _gen_integer(n):
    return np.random.randint(-100, 100, n)


def _gen_string(n):
    # Slightly hacks, but for testing purposes we'll ensure that we are generating
    # blobs that can be cast to other types as well.
    return list(str(x) for x in _gen_floating(n, low=0))


def _gen_blob(n):
    return list(x.encode("utf-8") for x in _gen_string(n))


def _gen_timestamp(n):
    start_time = np.datetime64('2017-01-01', 'ns')
    return np.array([(start_time + np.timedelta64(i, 's'))
                     for i in range(n)]).astype('datetime64[ns]')


def _sparsify(xs):
    # Randomly make a bunch of elements null, we make use of Numpy's masked
    # arrays for this: keep track of a separate boolean 'mask' array, which
    # determines whether or not an item in the array is None.
    mask = np.random.choice(a=[True, False], size=len(xs))
    return ma.masked_array(data=xs,
                           mask=mask)


@pytest.mark.parametrize("sparse", [True, False])
@pytest.mark.parametrize("write_fn", [qdbpd.write_dataframe,
                                      qdbpd.write_pinned_dataframe])
@pytest.mark.parametrize("input_gen", [_gen_floating,
                                       _gen_integer,
                                       _gen_blob,
                                       _gen_string,
                                       _gen_timestamp])
@pytest.mark.parametrize("column_type", [quasardb.ColumnType.Int64,
                                         quasardb.ColumnType.Double,
                                         quasardb.ColumnType.Blob,
                                         quasardb.ColumnType.String,
                                         quasardb.ColumnType.Timestamp])
def test_inference(
        write_fn,
        qdbd_connection,
        table_name,
        input_gen,
        column_type,
        sparse):

    # Create table
    t = qdbd_connection.ts(table_name)
    t.create([quasardb.ColumnInfo(column_type, "x")])

    n = 100
    idx = pd.date_range(np.datetime64('2017-01-01'), periods=n, freq='S')
    xs = input_gen(n)
    if sparse is True:
        xs = _sparsify(xs)

    df = pd.DataFrame(data={"x": xs}, index=idx)

    # Note that there are no tests; we effectively only test whether it doesn't
    # throw.
    write_fn(df, qdbd_connection, t)



import re
import pytest
import quasardb
import logging
import time
import numpy as np


def test_native_logging_output(blob_entry, random_string, caplog):
    caplog.set_level(logging.DEBUG)

    # Do something that generates native logs
    blob_entry.put(random_string)

    # Wait for QuasarDB to flush the logs
    time.sleep(6)

    # Do something that will flush the logs
    blob_entry.remove()

    # Gather all module names that have generated logs, we expect some stuff from at
    # least quasardb.native which we want to test here
    messages = list(lr.message for lr in caplog.records)
    modules = list(lr.name for lr in caplog.records)

    print("messages: ", messages)
    print("modules: ", modules)

    assert 'quasardb.native' in modules
    assert 'requested clients broker service run' in messages


def test_invalid_utf8_logs_qdb3361(qdbd_connection, caplog):
    caplog.set_level(logging.DEBUG)

    # Generate invalid UTF-8 binary string which we'll use as table name
    tablename = b'foo\x80abc'

    # Verify this does, in fact, raise an error in case when it's interpreted
    # as UTF-8.
    with pytest.raises(UnicodeDecodeError):
        tablename.decode("utf-8")

    # Now, use this tablename to do some batch insertion
    table = qdbd_connection.table(tablename)
    try:
        table.remove()
    except quasardb.Error:
        # AliasNotFound
        pass

    col = quasardb.ColumnInfo(quasardb.ColumnType.Double, "the_double")
    table.create([col])

    batchcol = [quasardb.BatchColumnInfo(tablename, "the_double", 10)]
    inserter = qdbd_connection.inserter(batchcol)

    inserter.start_row(np.datetime64('2020-01-01', 'ns'))
    inserter.set_double(0, 1.234)

    inserter.push()
    time.sleep(4)

    # This call right here triggers the 'bad' logs to be flushed into the python
    # userspace.
    table.remove()

    # Look at the captured log messages to verify our 'broken' utf-8 code is, in fact,
    # in the logs.
    seen = False
    for lr in caplog.records:
        seen = seen or '\x80' in lr.message

    assert seen



# pylint: disable=C0103,C0111,C0302,W0212
import datetime

import pytest
import quasardb


def test_build():
    build = quasardb.build()
    assert len(build) > 0


def test_version():
    build = quasardb.version()
    assert len(build) > 0


def test_can_purge_all(qdbd_secure_connection):
    qdbd_secure_connection.purge_all(datetime.timedelta(minutes=1))



import re
import pytest
import quasardb
import quasardb.pool as pool

def test_can_create_pool():
    # Basic test whether whether we can create a new pool

    with pool.SingletonPool(uri='qdb://127.0.0.1:2836') as p:
        pass

def test_can_acquire_connection():
    # Basic test whether whether we can acquire a new connection

    with pool.SingletonPool(uri='qdb://127.0.0.1:2836') as p:
        with p.connect() as conn:
            pass

def test_releases_connection_in_with_block():
    # Tests whether pool.release() is invoked with a connection after a `with` block,
    # rather than actually closing a connection

    the_conn = None

    with pool.SingletonPool(uri='qdb://127.0.0.1:2836') as p:
        with p.connect() as conn:
            the_conn = conn
            assert the_conn.is_open() is True

        assert the_conn.is_open() is True


def test_cleans_connections_after_pool_close():
    # Verifies whether a pool cleans up all connections after it is closed

    the_conn = None

    with pool.SingletonPool(uri='qdb://127.0.0.1:2836') as p:
        with p.connect() as conn:
            the_conn = conn
            assert the_conn.is_open() is True

        assert the_conn.is_open() is True

    assert the_conn.is_open() is False



def test_use_acquired_connection(entry_name):
    # Ensures the connection returned by `pool.connect()` is actually a usable
    # QuasarDB connection. It's relevant because we extend the connection class
    # for our SingletonPool.

    with pool.SingletonPool(uri='qdb://127.0.0.1:2836') as p:
        with p.connect() as conn:
            conn.integer(entry_name).put(1234)
            assert conn.integer(entry_name).get() == 1234

def test_uninitialized_singleton():
    # Tests whether an error is thrown if we use the singleton without initialization
    with pytest.raises(AssertionError, match=r"Global connection pool uninitialized"):
        pool.instance()

def test_singleton():
    pool.initialize(uri='qdb://127.0.0.1:2836')
    assert isinstance(pool.instance(), pool.SingletonPool)

def test_singleton_checks_type():
    # Tests whether an error is thrown if we initialize the pool with an incorrect
    # arg.
    with pytest.raises(TypeError):
        pool.initialize('Foo')

@pool.with_conn()
def _test_default_decorator_impl(conn):
    # Default decorator just inserts the pool as the first arg
    return conn

def test_default_decorator():
    assert isinstance(_test_default_decorator_impl(), pool.SessionWrapper)

@pool.with_conn(arg='foo')
def _test_kwargs_decorator_impl(entry_name, foo):
    # We can use arg= to specify a specific kwarg that should be used
    return foo

def test_kwarg_decorator(entry_name):
    assert isinstance(_test_kwargs_decorator_impl(entry_name), pool.SessionWrapper)

@pool.with_conn(arg=1)
def _test_positional_args_decorator_impl(entry_name, conn, another_entry_name):
    # We expect the pool to be injected at exactly position 1
    return conn

def test_positional_args_decorator(entry_name):
    assert isinstance(_test_positional_args_decorator_impl(entry_name, entry_name), pool.SessionWrapper)



# pylint: disable=C0103,C0111,C0302,R0903

# Copyright (c) 2009-2021, quasardb SAS. All rights reserved.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of quasardb nor the names of its contributors may
#      be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY QUASARDB AND CONTRIBUTORS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import quasardb
import logging
import time
from datetime import datetime
from functools import partial


logger = logging.getLogger('quasardb.pandas')


class PandasRequired(ImportError):
    """
    Exception raised when trying to use QuasarDB pandas integration, but
    pandas has not been installed.
    """
    pass


try:
    import numpy as np
    import pandas as pd
    from pandas.core.api import DataFrame, Series
    from pandas.core.base import PandasObject
except ImportError:
    raise PandasRequired(
        "The pandas library is required to handle pandas data formats")


# Constant mapping of numpy dtype to QuasarDB column type
# TODO(leon): support this natively in qdb C api ? we have everything we need
#             to understand dtypes.
_dtype_map = {
    np.dtype('int64'): quasardb.ColumnType.Int64,
    np.dtype('int32'): quasardb.ColumnType.Int64,
    np.dtype('float64'): quasardb.ColumnType.Double,
    np.dtype('object'): quasardb.ColumnType.String,
    np.dtype('M8[ns]'): quasardb.ColumnType.Timestamp,
    np.dtype('datetime64[ns]'): quasardb.ColumnType.Timestamp,

    'int64': quasardb.ColumnType.Int64,
    'int32': quasardb.ColumnType.Int64,
    'float32': quasardb.ColumnType.Double,
    'float64': quasardb.ColumnType.Double,
    'timestamp': quasardb.ColumnType.Timestamp,
    'string': quasardb.ColumnType.String,
    'bytes': quasardb.ColumnType.Blob,

    'floating': quasardb.ColumnType.Double,
    'integer': quasardb.ColumnType.Int64,
    'bytes': quasardb.ColumnType.Blob,
    'string': quasardb.ColumnType.String,
    'datetime64':  quasardb.ColumnType.Timestamp
}

# Based on QuasarDB column types, which dtype do we want?
_dtypes_map_flip = {
    quasardb.ColumnType.String: np.dtype('unicode'),
    quasardb.ColumnType.Int64: np.dtype('int64'),
    quasardb.ColumnType.Double: np.dtype('float64'),
    quasardb.ColumnType.Blob: np.dtype('object'),
    quasardb.ColumnType.Timestamp: np.dtype('datetime64[ns]')
}


def read_series(table, col_name, ranges=None):
    """
    Read a Pandas Timeseries from a single column.

    Parameters:
    -----------

    table : quasardb.Timeseries
      QuasarDB Timeseries table object, e.g. qdb_cluster.table('my_table')

    col_name : str
      Name of the column to read.

    ranges : list
      A list of ranges to read, represented as tuples of Numpy datetime64[ns] objects.
    """
    read_with = {
        quasardb.ColumnType.Double: table.double_get_ranges,
        quasardb.ColumnType.Blob: table.blob_get_ranges,
        quasardb.ColumnType.String: table.string_get_ranges,
        quasardb.ColumnType.Int64: table.int64_get_ranges,
        quasardb.ColumnType.Timestamp: table.timestamp_get_ranges,
    }

    kwargs = {
        'column': col_name
    }

    if ranges is not None:
        kwargs['ranges'] = ranges

    # Dispatch based on column type
    t = table.column_type_by_id(col_name)

    logger.debug(
        "reading Series from column %s.%s with type %s",
        table.get_name(),
        col_name,
        t)

    res = (read_with[t])(**kwargs)

    return Series(res[1], index=res[0])


def write_series(series, table, col_name):
    """
    Writes a Pandas Timeseries to a single column.

    Parameters:
    -----------

    series : pandas.Series
      Pandas Series, with a numpy.datetime64[ns] as index. Underlying data will be attempted
      to be transformed to appropriate QuasarDB type.

    table : quasardb.Timeseries
      QuasarDB Timeseries table object, e.g. qdb_cluster.table('my_table')

    col_name : str
      Column name to store data in.
    """
    write_with = {
        quasardb.ColumnType.Double: table.double_insert,
        quasardb.ColumnType.Blob: table.blob_insert,
        quasardb.ColumnType.String: table.string_insert,
        quasardb.ColumnType.Int64: table.int64_insert,
        quasardb.ColumnType.Timestamp: table.timestamp_insert
    }

    t = table.column_type_by_id(col_name)

    xs = series.to_numpy(_dtypes_map_flip[t])
    logger.debug(
        "writing Series to column %s.%s with type %s",
        table.get_name(),
        col_name,
        t)

    (write_with[t])(col_name, series.index.to_numpy(), xs)


def query(cluster, query, blobs=False):
    """
    Execute a query and return the results as DataFrames. Returns a dict of
    tablename / DataFrame pairs.

    Parameters:
    -----------

    cluster : quasardb.Cluster
      Active connection to the QuasarDB cluster

    query : str
      The query to execute.

    blobs : bool or list
      Determines which QuasarDB blob-columns should be returned as bytearrays; otherwise
      they are returned as UTF-8 strings.

      True means every blob column should be returned as byte-array, or a list will
      specify which specific columns. Defaults to false, meaning all blobs are returned
      as strings.

    """
    logger.debug("querying and returning as DataFrame: %s", query)

    return DataFrame(cluster.query(query, blobs=blobs))


def read_dataframe(table, row_index=False, columns=None, ranges=None):
    """
    Read a Pandas Dataframe from a QuasarDB Timeseries table.

    Parameters:
    -----------

    table : quasardb.Timeseries
      QuasarDB Timeseries table object, e.g. qdb_cluster.table('my_table')

    columns : optional list
      List of columns to read in dataframe. The timestamp column '$timestamp' is
      always read.

      Defaults to all columns.

    row_index: boolean
      Whether or not to index by rows rather than timestamps. Set to true if your
      dataset may contains null values and multiple rows may use the same timestamps.
      Note: using row_index is significantly slower.
      Defaults to false.

    ranges: optional list
      A list of time ranges to read, represented as tuples of Numpy datetime64[ns] objects.
      Defaults to the entire table.

    """

    if columns is None:
        columns = list(c.name for c in table.list_columns())

    if not row_index:
        logger.debug("reading DataFrame from %d series", len(columns))
        xs = dict((c, (read_series(table, c, ranges))) for c in columns)
        return DataFrame(data=xs)

    kwargs = {
        'columns': columns
    }
    if ranges:
        kwargs['ranges'] = ranges

    logger.debug("reading DataFrame from bulk reader")

    reader = table.reader(**kwargs)
    xs = []
    for row in reader:
        xs.append(row.copy())

    columns.insert(0, '$timestamp')

    logger.debug(
        "read %d rows, returning as DataFrame with %d columns",
        len(xs),
        len(columns))

    return DataFrame(data=xs, columns=columns)


def string_to_timestamp(x):
    return np.datetime64(datetime.utcfromtimestamp(int(float(x))), 'ns')


def fnil(f, x):
    "Utility function, only apply f to x if x is not nillable"

    if pd.isnull(x):
        return None

    if x == b'nan' or x == 'nan':
        return None

    return f(x)


_infer_with = {
    quasardb.ColumnType.Int64: {
        'floating': np.int64,
        'integer': np.int64,
        'string': lambda x: np.float64(x).astype(np.int64),
        'bytes': lambda x: np.float64(x.decode("utf-8")).astype(np.int64),
        'datetime64': lambda x: np.int64(x.nanosecond),
        'datetime': lambda x: np.int64(x.timestamp() * 1000000000),
        '_': lambda x: np.float64(x).astype(np.int64)
    },
    quasardb.ColumnType.Double: {
        'floating': lambda x: x,
        'integer': np.float64,
        'string': np.float64,
        'bytes': lambda x: np.float64(x.decode("utf-8")),
        'datetime64': lambda x: np.float64(x.nanosecond),
        '_': np.float64
    },
    quasardb.ColumnType.Blob: {
        'floating': lambda x: str(x).encode("utf-8"),
        'integer': lambda x: str(x).encode("utf-8"),
        'string': lambda x: str(x).encode("utf-8"),
        'bytes': lambda x: x,
        '_': lambda x: str(x).encode("utf-8"),
    },
    quasardb.ColumnType.String: {
        'floating': str,
        'integer': str,
        'string': str,
        'bytes': lambda x: x.decode("utf-8"),
        '_': str
    },
    quasardb.ColumnType.Timestamp: {
        'floating': lambda x: np.datetime64(datetime.utcfromtimestamp(x), 'ns'),
        'integer': lambda x: np.datetime64(datetime.utcfromtimestamp(x), 'ns'),
        'string': string_to_timestamp,
        'bytes': lambda x: string_to_timestamp(x.decode("utf-8")),
        'datetime64': lambda x: np.datetime64(x, 'ns'),
        '_': lambda x: np.datetime64(datetime.utcfromtimestamp(np.float64(x)), 'ns')
    }
}


# Update all function pointers to wrap them in _fnil, so that we don't
# accidentally convert e.g. None into string 'None'
for ct in _infer_with:
    for dt in _infer_with[ct]:
        f = _infer_with[ct][dt]
        _infer_with[ct][dt] = partial(fnil, f)


def write_dataframe(
        df,
        cluster,
        table,
        shard_size=None,
        create=False,
        _async=False,
        fast=False,
        truncate=False,
        infer_types=True):
    """
    Store a dataframe into a table.

    Parameters:
    df: pandas.DataFrame
      The pandas dataframe to store.

    cluster: quasardb.Cluster
      Active connection to the QuasarDB cluster

    table: quasardb.Timeseries or str
      Either a string or a reference to a QuasarDB Timeseries table object.
      For example, 'my_table' or cluster.table('my_table') are both valid values.

    create: bool, default False
      Whether to create the table. Defaults to false.

    shard_size: optional bool
      The shard size of the timeseries you wish to create.

    infer_types: bool, default True
      If true, will attemp to convert types from Python to QuasarDB natives types if
      the provided dataframe has incompatible types. For example, a dataframe with integers
      will automatically convert these to doubles if the QuasarDB table expects it.

      Defaults to True. For production use cases where you want to avoid implicit conversions,
      we recommend setting this to False.

    truncate: bool, default False
      Truncate (also referred to as upsert) the data in-place. Will detect time range to truncate
      from the time range inside the dataframe.

      Defaults to False.

    _async: bool, default False
      If true, uses asynchronous insertion API where commits are buffered server-side and
      acknowledged before they are written to disk. If you insert to the same table from
      multiple processes, setting this to True may improve performance.

      Defaults to False.

    fast: bool, default False
      Whether to use 'fast push'. If you incrementally add small batches of data to table,
      you may see better performance if you set this to True.

      Defaults to False.

    """

    # Acquire reference to table if string is provided
    if isinstance(table, str):
        table = cluster.table(table)

    if create:
        _create_table_from_df(df, table, shard_size)

    # Create batch column info from dataframe
    col_info = list(quasardb.BatchColumnInfo(
        table.get_name(), c, df.shape[0]) for c in df.columns)
    batch = cluster.inserter(col_info)

    write_with = {
        quasardb.ColumnType.Double: batch.set_double,
        quasardb.ColumnType.Blob: batch.set_blob,
        quasardb.ColumnType.String: batch.set_string,
        quasardb.ColumnType.Int64: batch.set_int64,
        quasardb.ColumnType.Timestamp: lambda i,
        x: batch.set_timestamp(
            i,
            np.datetime64(
                x,
                'ns'))}

    # We derive our column types from our table.
    ctypes = dict()
    for c in table.list_columns():
        ctypes[c.name] = c.type
    # Performance improvement: avoid a expensive dict lookups by indexing
    # the column types by relative offset within the df.
    ctypes_indexed = list(ctypes[c] for c in df.columns)

    dtypes = dict()
    dtypes_indexed = list()
    if infer_types is True:
        for i in range(len(df.columns)):
            c = df.columns[i]
            dt = pd.api.types.infer_dtype(df[c].values)
            logger.debug("Determined dtype of column %s to be %s", c, dt)
            dtypes[c] = dt

        # Performance improvement: avoid a expensive dict lookups by indexing
        # the column types by relative offset within the df.
        dtypes_indexed = list(dtypes[c] for c in df.columns)

    # TODO(leon): use pinned columns so we can write entire numpy arrays
    for row in df.itertuples(index=True):
        if pd.isnull(row[0]):
            raise RuntimeError(
                "Index must be a valid timestamp, found: " + str(row[0]))

        batch.start_row(np.datetime64(row[0], 'ns'))

        for i in range(len(df.columns)):
            v = row[i + 1]
            ct = ctypes_indexed[i]

            if infer_types is True:
                dt = dtypes_indexed[i]
                try:
                    fn = _infer_with[ct][dt]
                except KeyError:
                    # Fallback default
                    fn = _infer_with[ct]['_']

                v = fn(v)

            if not pd.isnull(v) and not pd.isna(v):

                fn = write_with[ct]

                try:
                    fn(i, v)
                except TypeError:
                    logger.exception(
                        "An error occured while setting column value: %s = %s", df.columns[i], v)
                    raise
                except ValueError:
                    logger.exception(
                        "An error occured while setting column value: %s = %s", df.columns[i], v)
                    raise

    start = time.time()

    logger.debug(
        "push chunk of %d rows, fast?=%s, async?=%s", len(
            df.index), fast, _async)

    if fast is True:
        batch.push_fast()
    elif truncate is True:
        batch.push_truncate()
    elif isinstance(truncate, tuple):
        batch.push_truncate(range=truncate)
    elif _async is True:
        batch.push_async()
    else:
        batch.push()

    logger.debug(
        "pushed %d rows in %s seconds", len(
            df.index), (time.time() - start))


def write_pinned_dataframe(
        df,
        cluster,
        table,
        create=False,
        shard_size=None,
        _async=False,
        fast=False,
        truncate=False,
        infer_types=True):
    """
    Store a dataframe into a table with the pin column API.
    Parameters:
    -----------

    df: pandas.DataFrame
      The pandas dataframe to store.
    cluster: quasardb.Cluster
      Active connection to the QuasarDB cluster
    table: quasardb.Timeseries or str
      Either a string or a reference to a QuasarDB Timeseries table object.
      For example, 'my_table' or cluster.table('my_table') are both valid values.
    create: optional bool
      Whether to create the table. Defaults to false.
    shard_size: optional bool
      The shard size of the timeseries you wish to create.
    infer_types: optional bool
      If true, will attemp to convert types from Python to QuasarDB natives types if
      the provided dataframe has incompatible types. For example, a dataframe with integers
      will automatically convert these to doubles if the QuasarDB table expects it.
      Defaults to True. For production use cases where you want to avoid implicit conversions,
      we recommend setting this to False.
    truncate: optional bool
      Truncate (also referred to as upsert) the data in-place. Will detect time range to truncate
      from the time range inside the dataframe.
      Defaults to False.
    _async: optional bool
      If true, uses asynchronous insertion API where commits are buffered server-side and
      acknowledged before they are written to disk. If you insert to the same table from
      multiple processes, setting this to True may improve performance.
      Defaults to False.
    fast: optional bool
      Whether to use 'fast push'. If you incrementally add small batches of data to table,
      you may see better performance if you set this to True.
      Defaults to False.
    """

    if not df.index.is_monotonic_increasing:
        logger.warn("dataframe index is unsorted, resorting dataframe based on index")
        df = df.sort_index().reindex()

    # Acquire reference to table if string is provided
    if isinstance(table, str):
        table = cluster.table(table)

    if create:
        _create_table_from_df(df, table, shard_size)

    # Create batch column info from dataframe
    writer = cluster.pinned_writer([table])

    write_with = {
        quasardb.ColumnType.Double: writer.set_double_column,
        quasardb.ColumnType.Blob: writer.set_blob_column,
        quasardb.ColumnType.String: writer.set_string_column,
        quasardb.ColumnType.Int64: writer.set_int64_column,
        quasardb.ColumnType.Timestamp: writer.set_timestamp_column,
    }

    pin_dtypes_map_flip = {
        quasardb.ColumnType.String: np.dtype('unicode'),
        quasardb.ColumnType.Int64: np.dtype('int64'),
        quasardb.ColumnType.Double: np.dtype('float64'),
        quasardb.ColumnType.Blob: np.dtype('object'),
        quasardb.ColumnType.Timestamp: np.dtype('datetime64[ns]')
    }

    # We derive our column types from our table.
    batch_columns = writer.batch_column_infos()
    column_types = writer.column_types()
    table_name = table.get_name()

    # Reorder & introduce new columns
    # https://stackoverflow.com/questions/13148429/how-to-change-the-order-of-dataframe-columns
    cnames = []
    ctypes = []
    for i in range(len(batch_columns)):
        bc = batch_columns[i]
        cnames.append(bc.column)
        ctypes.append(column_types[i])
        # Introduce new column
        found = False
        for c in df.columns:
            if table_name == bc.timeseries and c == bc.column:
                found = True
        if found == False:
            df.insert(i, bc.column, None)
    # Reorder columns
    df = df[cnames]

    dtypes = _get_inferred_dtypes_indexed(df) if infer_types is True else list()

    for i in range(len(df.columns)):
        ct = ctypes[i]
        tmp = df.iloc[:, i]
        timestamps = tmp.index.astype(np.dtype('int64')).tolist()
        values = []
        if infer_types is True:
            dt = dtypes[i]
            fn = None
            try:
                fn = _infer_with[ct][dt]
            except KeyError:
                # Fallback default
                fn = _infer_with[ct]['_']
            values = tmp.apply(lambda v: fn(v) if not pd.isnull(v) and not pd.isna(v) else None).tolist()
        else:
            dt = pin_dtypes_map_flip[ct]
            if (ct == quasardb.ColumnType.Int64):
                # you need to fill with 0x8000000000000000 (which is quasardb value for null)
                # astype(np.dtype('int64')) cannot convert None to int64
                values = tmp.fillna(0x8000000000000000).astype(dt).tolist()
            elif (ct == quasardb.ColumnType.Blob):
                values = tmp.fillna(b'').astype(dt).tolist()
            elif (ct == quasardb.ColumnType.String):
                values = tmp.fillna('').astype(dt).tolist()
            else:
                values = tmp.astype(dt).tolist()
        write_with[ct](i, timestamps, values)

    start = time.time()

    logger.debug("push chunk of %d rows, fast?=%s, async?=%s",
                 len(df.index), fast, _async)

    if fast is True:
        writer.push_fast()
    elif truncate is True:
        writer.push_truncate()
    elif isinstance(truncate, tuple):
        writer.push_truncate(range=truncate)
    elif _async is True:
        writer.push_async()
    else:
        writer.push()

    logger.debug(
        "pushed %d rows in %s seconds", len(
            df.index), (time.time() - start))

    return table


def _create_table_from_df(df, table, shard_size=None):
    cols = list()

    dtypes = _get_inferred_dtypes(df)
    for c in df.columns:
        dt = dtypes[c]
        ct = _dtype_to_column_type(df[c].dtype, dt)
        logger.debug("probed pandas dtype %s to inferred dtype %s and map to quasardb column type %s", df[c].dtype, dt, ct)
        cols.append(quasardb.ColumnInfo(ct, c))


    try:
        if not shard_size:
            table.create(cols)
        else:
            table.create(cols, shard_size)
    except quasardb.quasardb.AliasAlreadyExistsError:
        # TODO(leon): warn? how?
        pass

    return table

def _dtype_to_column_type(dt, inferred):
    res = _dtype_map.get(inferred, None)
    if res is None:
        res = _dtype_map.get(dt, None)

    if res is None:
        raise ValueError("Incompatible data type: ", dt)

    return res

def _get_inferred_dtypes(df):
    dtypes = dict()
    for i in range(len(df.columns)):
        c = df.columns[i]
        dt = pd.api.types.infer_dtype(df[c].values)
        logger.debug("Determined dtype of column %s to be %s", c, dt)
        dtypes[c] = dt
    return dtypes

def _get_inferred_dtypes_indexed(df):
    dtypes = _get_inferred_dtypes(df)
    # Performance improvement: avoid a expensive dict lookups by indexing
    # the column types by relative offset within the df.
    return list(dtypes[c] for c in df.columns)



# pylint: disable=C0103,C0111,C0302,W0212
import unittest
import pytest

import quasardb


def test_connect_with_trace():
    with pytest.raises(quasardb.Error, match=r"at qdb_connect"):
        quasardb.Cluster(uri='invalid_uri')



#!/usr/bin/env python
# (c)quasardb SAS. All rights reserved.
# qdb is a trademark of quasardb SAS

# pylint: disable=C0103,C0111,C0326,W0201


import os
import re
import sys
import platform
import subprocess
import glob

from distutils.version import LooseVersion
from setuptools.command.build_ext import build_ext

from setuptools import setup, Extension
from setuptools.command.bdist_egg import bdist_egg as old_bdist_egg  # pylint: disable=C0412
from pkg_resources import get_build_platform
from wheel.bdist_wheel import bdist_wheel as old_bdist_wheel

qdb_version = "3.10.2"

# package_modules are our 'extra' files. Our cmake configuration copies our QDB_API_LIB
# into our source directory, and by adding this to `package_modules` we tell setuptools to
# package this.
package_modules = glob.glob(os.path.join('quasardb', 'lib*'))
package_name = 'quasardb'


class CMakeExtension(Extension):
    def __init__(self, name, sourcedir=''):
        Extension.__init__(self, name, sources=[])
        self.sourcedir = os.path.abspath(sourcedir)


class CMakeBuild(build_ext):
    def run(self):
        try:
            out = subprocess.check_output(['cmake', '--version'])
        except OSError:
            raise RuntimeError("CMake must be installed to build the following extensions: " +
                               ", ".join(e.name for e in self.extensions))

        if platform.system() == "Windows":
            cmake_version = LooseVersion(
                re.search(r'version\s*([\d.]+)', out.decode()).group(1))
            if cmake_version < '3.1.0':
                raise RuntimeError("CMake >= 3.1.0 is required on Windows")

        for ext in self.extensions:
            self.build_extension(ext)

    def build_extension(self, ext):
        extdir = os.path.join(
            os.path.abspath(os.path.dirname(self.get_ext_fullpath(ext.name))), 'quasardb')

        # We provide CMAKE_LIBRARY_OUTPUT_DIRECTORY to cmake, where it will copy libqdb_api.so (or
        # whatever the OS uses). It is important that this path matches `package_modules`, so that
        # setuptools knows it needs to package this.
        cmake_args = ['-DCMAKE_LIBRARY_OUTPUT_DIRECTORY=' + extdir,
                      '-DPYTHON_EXECUTABLE=' + sys.executable,
                      '-DQDB_PY_VERSION=' + qdb_version]

        cfg = 'Debug' if self.debug else 'Release'
        build_args = ['--config', cfg]

        if platform.system() == "Windows":
            cmake_args += [
                '-DCMAKE_LIBRARY_OUTPUT_DIRECTORY_{}={}'.format(cfg.upper(), extdir)]
            cmake_args += ['-T', 'host=x64']
            if sys.maxsize > 2**32:
                cmake_args += ['-A', 'x64']
            else:
                cmake_args += ['-A', 'Win32']
        #    build_args += ['--', '/m']
        else:
            cmake_args += ['-DCMAKE_BUILD_TYPE=' + cfg]
        #    build_args += ['--', '-j2']

        env = os.environ.copy()
        env['CXXFLAGS'] = '{} -DVERSION_INFO=\\"{}\\"'.format(env.get('CXXFLAGS', ''),
                                                              self.distribution.get_version())
        if not os.path.exists(self.build_temp):
            os.makedirs(self.build_temp)
        subprocess.check_call(['cmake', ext.sourcedir] +
                              cmake_args, cwd=self.build_temp, env=env)
        subprocess.check_call(['cmake', '--build', '.'] +
                              build_args, cwd=self.build_temp)

# Generated by CMake
class EggRetagger(old_bdist_egg):
    def finalize_options(self):
        if self.plat_name is None:
            self.plat_name = get_build_platform()

        old_bdist_egg.finalize_options(self)

# Generated by CMake
class WheelRetagger(old_bdist_wheel):
    def get_tag(self):
        tag = old_bdist_wheel.get_tag(self)

        python_tag = tag[0]
        abi_tag = tag[1]
        platform_tag = tag[2]

        if platform_tag.startswith('macosx_10_') and platform_tag.endswith('_x86_64'):
            supported_versions = [
                'macosx_10_6', 'macosx_10_9', 'macosx_10_10', 'macosx_10_11', 'macosx_10_12']
            supported_versions = [
                version + '_x86_64' for version in supported_versions]
            platform_tag = '.'.join(supported_versions)

        tag = (python_tag, abi_tag, platform_tag)
        return tag

cmdclass = {
    'build_ext': CMakeBuild,
    'bdist_egg': EggRetagger,
    'bdist_wheel': WheelRetagger,
}

setup(name=package_name,
      version=qdb_version,
      description='Python API for quasardb',
      author='quasardb SAS',
      license='BSD',
      author_email='contact@quasardb.net',
      url='https://www.quasardb.net/',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: End Users/Desktop',
          'Intended Audience :: Financial and Insurance Industry',
          'Intended Audience :: Information Technology',
          'Intended Audience :: Other Audience',
          'Intended Audience :: System Administrators',
          'Intended Audience :: Telecommunications Industry',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Topic :: Database',
          'Topic :: Software Development :: Libraries :: Python Modules',
          "License :: OSI Approved :: BSD License",
      ],
      keywords='quasardb timeseries database API driver ',
      setup_requires=["setuptools-git"],
      install_requires=["numpy"],
      extras_require={
          "pandas": ["pandas"],
          "tests": [
              "pytest >= 6.2.5",
              "pytest-runner >= 5.3.1",
              "pytest-benchmark == 3.4.1",
              "teamcity-messages >= 1.29"]
      },

      packages=[package_name],
      package_data={package_name: package_modules},
      ext_modules=[CMakeExtension('quasardb', 'quasardb/')],
      include_package_data=True,
      cmdclass=cmdclass,
      zip_safe=False,
      test_suite="tests",
      )



# pylint: disable=C0103,C0111,C0302,W0212

import quasardb
import pytest

import test_batch_inserter as batchlib


def test_get_client_max_parallelism(qdbd_connection):
    assert qdbd_connection.options().get_client_max_parallelism() > 0


def test_set_client_max_parallelism_negative(qdbd_connection):
    with pytest.raises(TypeError):
        qdbd_connection.options().set_client_max_parallelism(-1)


def test_set_client_max_parallelism_0(qdbd_connection):
    qdbd_connection.options().set_client_max_parallelism(0)
    assert qdbd_connection.options().get_client_max_parallelism() > 0


def test_set_client_max_parallelism_1(qdbd_connection):
    qdbd_connection.options().set_client_max_parallelism(1)
    assert qdbd_connection.options().get_client_max_parallelism() == 1


def test_get_client_max_parallelism_8(qdbd_connection):
    assert qdbd_connection.options().get_client_max_parallelism() == 1



# import-start
import quasardb
import quasardb.pandas as qdbpd

import pandas as pd
import numpy as np
# import-end

with quasardb.Cluster("qdb://127.0.0.1:2836") as c:

    # batch-insert-start

    # Prepare the entire DataFrame which we wish to store
    data = {'open': [3.40, 3.50],
            'close': [3.50, 3.55],
            'volume': [10000, 7500]}
    timestamps = np.array([np.datetime64('2019-02-01'),
                           np.datetime64('2019-02-02')],
                          dtype='datetime64[ns]')
    df = pd.DataFrame(data=data,
                      index=timestamps)

    # By providing the create=True parameter, we explicitly tell the Pandas connector to
    # create a new table based on the schema
    # of the DataFrame.
    qdbpd.write_dataframe(df, c, "stocks", create=True)

    # batch-insert-end


    # bulk-read-start

    ranges = [(np.datetime64('2019-02-01', 'ns'), np.datetime64('2019-02-02', 'ns'))]
    t = c.table("stocks")

    # By providing the row_index=True parameter, we explicitly tell the Pandas connector
    # to retrieve rows using the row-oriented bulk reader and use the default Pandas sequential
    # row index. This is useful if your dataset is sparse and may contain null values.
    df = qdbpd.read_dataframe(t, row_index=True, ranges=ranges)

    # bulk-read-end

    # column-insert-start

    # The Pandas connector maps QuasarDB's column-oriented API to Pandas Series. As such,
    # we initialize three Series for each of our three columns.
    #
    # Seperately, we generate a numpy array of timestamps. Since our three columns share
    # the same timestamps, we can reuse this as the index for all our Series.
    timestamps = np.array([np.datetime64('2019-02-01'), np.datetime64('2019-02-02')], dtype='datetime64[ns]')

    opens = pd.Series(data=[3.40, 3.50], index=timestamps, dtype=np.float64)
    closes = pd.Series(data=[3.50, 3.55], index=timestamps, dtype=np.float64)
    volumes = pd.Series(data=[10000, 7500], index=timestamps, dtype=np.int64)

    # We use the write_series function to insert column by column.
    qdbpd.write_series(opens, t, "open")
    qdbpd.write_series(closes, t, "close")
    qdbpd.write_series(volumes, t, "volume")

    # column-insert-end

    # column-get-start

    # We first prepare the intervals we want to select data from, that is, a list of
    # timeranges. An interval is defined as a tuple of start time (inclusive) and end
    # time (exclusive).
    #
    # In this example, we just use a single interval.
    intervals = np.array([(np.datetime64('2019-02-01', 'ns'), np.datetime64('2019-02-02', 'ns'))])

    # We can then use the read_series function to read column by column. The objects
    # returned are regular pd.Series objects.
    opens = qdbpd.read_series(t, "open", intervals)
    closes = qdbpd.read_series(t, "close", intervals)
    volumes = qdbpd.read_series(t, "volume", intervals)

    # column-get-end

    # query-start

    df = qdbpd.query(c, "SELECT SUM(volume) FROM stocks")

    # The API returns dataframe
    print("result: ", df)

    # query-end

    c.table("stocks").remove()



# pylint: disable=C0103,C0111,C0302,W0212
from builtins import range as xrange, int as long  # pylint: disable=W0622
from functools import reduce  # pylint: disable=W0622
import datetime
import test_table as tslib
from time import sleep

import pytest
import quasardb
import numpy as np


def _generate_data(count, start=np.datetime64('2017-01-01', 'ns')):
    integers = np.random.randint(-100, 100, count)
    timestamps = tslib._generate_dates(
        start + np.timedelta64('1', 'D'), count)

    return (integers, timestamps)


def test_rows_with_none_values(qdbd_connection, table_name):
    table = qdbd_connection.table(table_name)
    node_id = quasardb.ColumnInfo(quasardb.ColumnType.Int64, "node_id")
    x = quasardb.ColumnInfo(quasardb.ColumnType.Int64, "x")
    y = quasardb.ColumnInfo(quasardb.ColumnType.Int64, "y")

    table.create([node_id, x, y])

    pinned_writer = qdbd_connection.pinned_writer([table])

    timestamps = [
        np.datetime64(
            '2020-01-01T00:00:00',
            'ns'),
        np.datetime64(
            '2020-01-01T00:00:00',
            'ns')]
    node_id = [10, 12]
    x = [None, 1]
    y = [2, None]

    pinned_writer.set_int64_column(0, timestamps, node_id)
    pinned_writer.set_int64_column(1, timestamps, x)
    pinned_writer.set_int64_column(2, timestamps, y)

    pinned_writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","node_id","x","y" FROM "{}"'.format(
            table.get_name()))

    assert len(res) == 2
    for idx, row in enumerate(res):
        assert row['$timestamp'] == timestamps[idx]
        assert row['node_id'] == node_id[idx]
        assert row['x'] == x[idx]
        assert row['y'] == y[idx]


def test_incorrect_type_double(qdbd_connection, table):
    pinned_writer = qdbd_connection.pinned_writer([table])
    pinned_writer.start_row(np.datetime64('2020-01-01T00:00:00', 'ns'))
    for idx in range(5):
        if idx == 0:
            continue
        with pytest.raises(quasardb.Error):
            pinned_writer.set_double(idx, 1.1)


def test_successful_type_double(qdbd_connection, table):
    timestamp = np.datetime64('2020-01-01T00:00:00', 'ns')
    value = 1.1

    pinned_writer = qdbd_connection.pinned_writer([table])
    pinned_writer.start_row(timestamp)
    pinned_writer.set_double(0, value)
    pinned_writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_double" FROM "{}"'.format(
            table.get_name()))
    assert len(res) == 1
    assert res[0]['$timestamp'] == timestamp
    assert res[0]['the_double'] == value


def test_incorrect_type_blob(qdbd_connection, table):
    pinned_writer = qdbd_connection.pinned_writer([table])
    pinned_writer.start_row(np.datetime64('2020-01-01T00:00:00', 'ns'))
    for idx in range(5):
        if idx == 1:
            continue
        with pytest.raises(quasardb.Error):
            pinned_writer.set_blob(idx, b'aaa')


def test_successful_type_blob(qdbd_connection, table):
    timestamp = np.datetime64('2020-01-01T00:00:00', 'ns')
    value = b'aaa'

    pinned_writer = qdbd_connection.pinned_writer([table])
    pinned_writer.start_row(timestamp)
    pinned_writer.set_blob(1, value)
    pinned_writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_blob" FROM "{}"'.format(
            table.get_name()))
    assert len(res) == 1
    assert res[0]['$timestamp'] == timestamp
    assert res[0]['the_blob'] == value


def test_incorrect_type_string(qdbd_connection, table):
    pinned_writer = qdbd_connection.pinned_writer([table])
    pinned_writer.start_row(np.datetime64('2020-01-01T00:00:00', 'ns'))
    for idx in range(5):
        if idx == 2:
            continue
        with pytest.raises(quasardb.Error):
            pinned_writer.set_string(idx, 'aaa')


def test_successful_type_string(qdbd_connection, table):
    timestamp = np.datetime64('2020-01-01T00:00:00', 'ns')
    value = 'aaa'

    pinned_writer = qdbd_connection.pinned_writer([table])
    pinned_writer.start_row(timestamp)
    pinned_writer.set_string(2, value)
    pinned_writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_string" FROM "{}"'.format(
            table.get_name()))
    assert len(res) == 1
    assert res[0]['$timestamp'] == timestamp
    assert res[0]['the_string'] == value


def test_incorrect_type_int64(qdbd_connection, table):
    pinned_writer = qdbd_connection.pinned_writer([table])
    pinned_writer.start_row(np.datetime64('2020-01-01T00:00:00', 'ns'))
    for idx in range(5):
        if idx == 3:
            continue
        with pytest.raises(quasardb.Error):
            pinned_writer.set_int64(idx, 1)


def test_successful_type_int64(qdbd_connection, table):
    timestamp = np.datetime64('2020-01-01T00:00:00', 'ns')
    value = 1

    pinned_writer = qdbd_connection.pinned_writer([table])
    pinned_writer.start_row(timestamp)
    pinned_writer.set_int64(3, value)
    pinned_writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_int64" FROM "{}"'.format(
            table.get_name()))
    assert len(res) == 1
    assert res[0]['$timestamp'] == timestamp
    assert res[0]['the_int64'] == value


def test_incorrect_type_timestamp(qdbd_connection, table):
    pinned_writer = qdbd_connection.pinned_writer([table])
    pinned_writer.start_row(np.datetime64('2020-01-01T00:00:00', 'ns'))
    for idx in range(5):
        if idx == 4:
            continue
        with pytest.raises(quasardb.Error):
            pinned_writer.set_timestamp(
                idx, np.datetime64(
                    '2020-01-01T00:00:00', 'ns'))


def test_successful_type_timestamp(qdbd_connection, table):
    timestamp = np.datetime64('2020-01-01T00:00:00', 'ns')
    value = np.datetime64('2020-01-01T00:00:00', 'ns')

    pinned_writer = qdbd_connection.pinned_writer([table])
    pinned_writer.start_row(timestamp)
    pinned_writer.set_timestamp(4, value)
    pinned_writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_ts" FROM "{}"'.format(
            table.get_name()))
    assert len(res) == 1
    assert res[0]['$timestamp'] == timestamp
    assert res[0]['the_ts'] == value


def test_insert_data_ordered_single_shard(qdbd_connection, table):
    pinned_writer = qdbd_connection.pinned_writer([table])

    timestamps = [
        np.datetime64(
            '2020-01-01T00:00:00',
            'ns'),
        np.datetime64(
            '2020-01-01T00:00:01',
            'ns')]
    values = [0, 1]

    pinned_writer.start_row(timestamps[0])
    pinned_writer.set_int64(3, values[0])
    pinned_writer.start_row(timestamps[1])
    pinned_writer.set_int64(3, values[1])

    pinned_writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_int64" FROM "{}"'.format(
            table.get_name()))
    assert len(res) == 2
    for idx, row in enumerate(res):
        assert row['$timestamp'] == timestamps[idx]
        assert row['the_int64'] == values[idx]


def test_insert_data_ordered_multiple_shards(qdbd_connection, table):
    pinned_writer = qdbd_connection.pinned_writer([table])

    timestamps = [
        np.datetime64(
            '2020-01-01',
            'ns'),
        np.datetime64(
            '2020-01-02',
            'ns')]
    values = [1, 2]

    pinned_writer.start_row(timestamps[0])
    pinned_writer.set_int64(3, values[0])
    pinned_writer.start_row(timestamps[1])
    pinned_writer.set_int64(3, values[1])

    pinned_writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_int64" FROM "{}"'.format(
            table.get_name()))
    assert len(res) == 2
    for idx, row in enumerate(res):
        assert row['$timestamp'] == timestamps[idx]
        assert row['the_int64'] == values[idx]


def test_insert_data_unordered_single_shard(qdbd_connection, table):
    pinned_writer = qdbd_connection.pinned_writer([table])

    timestamps = [
        np.datetime64(
            '2020-01-01T00:00:01',
            'ns'),
        np.datetime64(
            '2020-01-01T00:00:00',
            'ns')]
    values = [1, 0]

    pinned_writer.start_row(timestamps[0])
    pinned_writer.set_int64(3, values[0])
    pinned_writer.start_row(timestamps[1])
    pinned_writer.set_int64(3, values[1])

    pinned_writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_int64" FROM "{}"'.format(
            table.get_name()))

    assert res[0]['$timestamp'] == timestamps[1]
    assert res[0]['the_int64'] == values[1]
    assert res[1]['$timestamp'] == timestamps[0]
    assert res[1]['the_int64'] == values[0]


def test_insert_data_unordered_multiple_shards(qdbd_connection, table):
    pinned_writer = qdbd_connection.pinned_writer([table])

    timestamps = [
        np.datetime64(
            '2020-01-02',
            'ns'),
        np.datetime64(
            '2020-01-01',
            'ns')]
    values = [2, 1]

    pinned_writer.start_row(timestamps[0])
    pinned_writer.set_int64(3, values[0])
    pinned_writer.start_row(timestamps[1])
    pinned_writer.set_int64(3, values[1])

    pinned_writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_int64" FROM "{}"'.format(
            table.get_name()))

    assert res[0]['$timestamp'] == timestamps[1]
    assert res[0]['the_int64'] == values[1]
    assert res[1]['$timestamp'] == timestamps[0]
    assert res[1]['the_int64'] == values[0]


def test_insert_data_ordered_single_shard_two_push(qdbd_connection, table):
    pinned_writer = qdbd_connection.pinned_writer([table])

    timestamps = [
        np.datetime64(
            '2020-01-01T00:00:00',
            'ns'),
        np.datetime64(
            '2020-01-01T00:00:01',
            'ns')]
    values = [0, 1]

    pinned_writer.start_row(timestamps[0])
    pinned_writer.set_int64(3, values[0])
    pinned_writer.push()

    pinned_writer.start_row(timestamps[1])
    pinned_writer.set_int64(3, values[1])
    pinned_writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_int64" FROM "{}"'.format(
            table.get_name()))
    assert len(res) == 2
    for idx, row in enumerate(res):
        assert row['$timestamp'] == timestamps[idx]
        assert row['the_int64'] == values[idx]


def test_insert_data_ordered_multiple_shards_two_push(qdbd_connection, table):
    pinned_writer = qdbd_connection.pinned_writer([table])

    timestamps = [
        np.datetime64(
            '2020-01-01',
            'ns'),
        np.datetime64(
            '2020-01-02',
            'ns')]
    values = [1, 2]

    pinned_writer.start_row(timestamps[0])
    pinned_writer.set_int64(3, values[0])
    pinned_writer.push()

    pinned_writer.start_row(timestamps[1])
    pinned_writer.set_int64(3, values[1])
    pinned_writer.push()

    res = qdbd_connection.query(
        'SELECT "$timestamp","the_int64" FROM "{}"'.format(
            table.get_name()))
    assert len(res) == 2
    for idx, row in enumerate(res):
        assert row['$timestamp'] == timestamps[idx]
        assert row['the_int64'] == values[idx]

# generative tests


def _row_insertion_method(
        writer,
        dates,
        doubles,
        blobs,
        strings,
        integers,
        timestamps):
    for i in range(len(dates)):
        writer.start_row(dates[i])
        writer.set_double(0, doubles[i])
        writer.set_blob(1, blobs[i])
        writer.set_string(2, strings[i])
        writer.set_int64(3, integers[i])
        writer.set_timestamp(4, timestamps[i])


def _regular_push(writer):
    writer.push()


def _async_push(writer):
    writer.push_async()
    # Wait for push_async to complete
    # Ideally we could be able to get the proper flush interval
    sleep(15)


def _fast_push(writer):
    writer.push_fast()


def _generate_data(count, start=np.datetime64('2017-01-01', 'ns')):
    doubles = np.random.uniform(-100.0, 100.0, count)
    integers = np.random.randint(-100, 100, count)
    blobs = np.array(list(np.random.bytes(np.random.randint(8, 16))
                          for i in range(count)), 'O')
    strings = np.array([("content_" + str(item)) for item in range(count)])
    timestamps = tslib._generate_dates(
        start + np.timedelta64('1', 'D'), count)

    return (doubles, integers, blobs, strings, timestamps)


def _set_batch_writer_data(writer, intervals, data, start=0):
    (doubles, integers, blobs, strings, timestamps) = data

    for i in range(start, len(intervals)):
        writer.start_row(intervals[i])
        writer.set_double(0, doubles[i])
        writer.set_blob(1, blobs[i])
        writer.set_string(2, strings[i])
        writer.set_int64(3, integers[i])
        writer.set_timestamp(4, timestamps[i])


def _assert_results(table, intervals, data):
    (doubles, integers, blobs, strings, timestamps) = data

    whole_range = (intervals[0], intervals[-1:][0] + np.timedelta64(2, 's'))
    results = table.double_get_ranges(
        tslib._double_col_name(table), [whole_range])

    np.testing.assert_array_equal(results[0], intervals)
    np.testing.assert_array_equal(results[1], doubles)

    results = table.blob_get_ranges(tslib._blob_col_name(table), [whole_range])
    np.testing.assert_array_equal(results[0], intervals)
    np.testing.assert_array_equal(results[1], blobs)

    results = table.string_get_ranges(
        tslib._string_col_name(table), [whole_range])
    np.testing.assert_array_equal(results[0], intervals)
    np.testing.assert_array_equal(results[1], strings)

    results = table.int64_get_ranges(
        tslib._int64_col_name(table), [whole_range])
    np.testing.assert_array_equal(results[0], intervals)
    np.testing.assert_array_equal(results[1], integers)

    results = table.timestamp_get_ranges(
        tslib._ts_col_name(table), [whole_range])
    np.testing.assert_array_equal(results[0], intervals)
    np.testing.assert_array_equal(results[1], timestamps)


def _test_with_table(
        writer,
        table,
        intervals,
        push_method=_regular_push,
        set_method=_set_batch_writer_data,
        data=None):

    if data is None:
        data = _generate_data(len(intervals))

    # range is right exclusive, so the timestamp has to be beyond
    whole_range = (intervals[0], intervals[-1:][0] + np.timedelta64(2, 's'))

    (doubles, integers, blobs, strings, timestamps) = data

    set_method(writer, intervals, data)

    # before the push, there is nothing
    results = table.double_get_ranges(
        tslib._double_col_name(table), [whole_range])
    assert len(results[0]) == 0

    results = table.blob_get_ranges(tslib._blob_col_name(table), [whole_range])
    assert len(results[0]) == 0

    results = table.string_get_ranges(
        tslib._string_col_name(table), [whole_range])
    assert len(results[0]) == 0

    results = table.int64_get_ranges(
        tslib._int64_col_name(table), [whole_range])
    assert len(results[0]) == 0

    results = table.timestamp_get_ranges(
        tslib._ts_col_name(table), [whole_range])
    assert len(results[0]) == 0

    # after push, there is everything
    push_method(writer)
    if push_method == _async_push:
        sleep(20)

    _assert_results(table, intervals, data)

    return doubles, blobs, strings, integers, timestamps


def test_insert(qdbd_connection, table, many_intervals):
    writer = qdbd_connection.pinned_writer([table])

    _test_with_table(
        writer,
        table,
        many_intervals,
        _regular_push)


def test_insert_secure(qdbd_secure_connection, secure_table, many_intervals):
    writer = qdbd_secure_connection.pinned_writer([secure_table])

    _test_with_table(
        writer,
        secure_table,
        many_intervals,
        _regular_push)


def test_insert_async(
        qdbd_connection, table, many_intervals):

    # Same test as `test_insert` but using `push_async` to push the entries
    # This allows us to test the `push_async` feature

    writer = qdbd_connection.pinned_writer([table])
    _test_with_table(
        writer,
        table,
        many_intervals,
        _async_push)


def test_insert_fast(
        qdbd_connection, table, many_intervals):
    # Same test as `test_insert` but using `push_fast` to push the entries
    # This allows us to test the `push_fast` feature

    writer = qdbd_connection.pinned_writer([table])
    _test_with_table(
        writer,
        table,
        many_intervals,
        _fast_push)


def test_insert_truncate_implicit_range(
        qdbd_connection, table, many_intervals):

    whole_range = (
        many_intervals[0], many_intervals[-1:][0] + np.timedelta64(2, 's'))

    # Generate our dataset
    data = _generate_data(len(many_intervals))
    # (doubles, integers, blobs, strings, timestamps) = data
    (doubles, _, _, _, _) = data

    # Insert once
    writer = qdbd_connection.pinned_writer([table])
    _set_batch_writer_data(writer, many_intervals, data)
    writer.push()

    # Compare results, should be equal
    results = table.double_get_ranges(
        tslib._double_col_name(table), [whole_range])

    np.testing.assert_array_equal(results[0], many_intervals)
    np.testing.assert_array_equal(results[1], doubles)

    # Insert regular, twice
    _set_batch_writer_data(writer, many_intervals, data)
    writer.push()

    # Compare results, should now have the same data twice
    results = table.double_get_ranges(
        tslib._double_col_name(table), [whole_range])

    assert len(results[1]) == 2 * len(doubles)

    # Insert truncate, should now have original data again
    _set_batch_writer_data(writer, many_intervals, data)
    writer.push_truncate()

    # Verify results, truncating should now make things the same
    # as the beginning again.
    results = table.double_get_ranges(
        tslib._double_col_name(table), [whole_range])

    np.testing.assert_array_equal(results[0], many_intervals)
    np.testing.assert_array_equal(results[1], doubles)


def test_insert_truncate_explicit_range(
        qdbd_connection, table, many_intervals):

    whole_range = (
        many_intervals[0], many_intervals[-1:][0] + np.timedelta64(2, 's'))

    # Generate our dataset
    data = _generate_data(len(many_intervals))
    # (doubles, integers, blobs, strings, timestamps) = data
    (doubles, _, _, _, _) = data

    writer = qdbd_connection.pinned_writer([table])

    # Insert once
    truncate_range = (whole_range[0],
                      whole_range[1] + np.timedelta64(1, 'ns'))

    _set_batch_writer_data(writer, many_intervals, data)
    writer.push()

    # Verify results, truncating should now make things the same
    # as the beginning again.
    results = table.double_get_ranges(
        tslib._double_col_name(table), [whole_range])

    np.testing.assert_array_equal(results[0], many_intervals)
    np.testing.assert_array_equal(results[1], doubles)

    # If we now set the same data, skip the first element, but keep
    # the same time range, the first element will *not* be present in
    # the resulting dataset.
    _set_batch_writer_data(writer, many_intervals, data, start=1)
    writer.push_truncate(range=truncate_range)

    # Verify results, truncating should now make things the same
    # as the beginning again.
    results = table.double_get_ranges(
        tslib._double_col_name(table), [whole_range])

    np.testing.assert_array_equal(results[0], many_intervals[1:])
    np.testing.assert_array_equal(results[1], doubles[1:])


def test_insert_truncate_throws_error_on_invalid_range(
        qdbd_connection, table, many_intervals):
    whole_range = (
        many_intervals[0], many_intervals[-1:][0] + np.timedelta64(2, 's'))

    # Generate our dataset
    data = _generate_data(len(many_intervals))
    # (doubles, integers, blobs, strings, timestamps) = data
    (_, _, _, _, _) = data

    # Insert truncate with explicit timerange, we point the start right after the
    # first element in our dataset. This means that the range does not overlap all
    # the data anymore.
    truncate_range = (whole_range[0] + np.timedelta64(1, 'ns'),
                      whole_range[1] + np.timedelta64(1, 'ns'))

    writer = qdbd_connection.pinned_writer([table])
    _set_batch_writer_data(writer, many_intervals, data)
    with pytest.raises(quasardb.Error):
        writer.push_truncate(range=truncate_range)


def _set_batch_writer_column_data(writer, intervals, data, start=0):
    (doubles, integers, blobs, strings, timestamps) = data

    writer.set_double_column(0, intervals[start:], doubles[start:])
    writer.set_blob_column(1, intervals[start:], blobs[start:])
    writer.set_string_column(2, intervals[start:], strings[start:])
    writer.set_int64_column(3, intervals[start:], integers[start:])
    writer.set_timestamp_column(4, intervals[start:], timestamps[start:])


def test_insert_column(qdbd_connection, table, many_intervals):
    writer = qdbd_connection.pinned_writer([table])

    _test_with_table(
        writer,
        table,
        many_intervals,
        _regular_push,
        _set_batch_writer_column_data)


def test_insert_column_secure(
        qdbd_secure_connection,
        secure_table,
        many_intervals):
    writer = qdbd_secure_connection.pinned_writer([secure_table])

    _test_with_table(
        writer,
        secure_table,
        many_intervals,
        _regular_push,
        _set_batch_writer_column_data)


def test_insert_column_async(
        qdbd_connection, table, many_intervals):

    # Same test as `test_insert_column` but using `push_async` to push the entries
    # This allows us to test the `push_async` feature

    writer = qdbd_connection.pinned_writer([table])
    _test_with_table(
        writer,
        table,
        many_intervals,
        _async_push,
        _set_batch_writer_column_data)


def test_insert_column_fast(
        qdbd_connection, table, many_intervals):
    # Same test as `test_insert_column` but using `push_fast` to push the entries
    # This allows us to test the `push_fast` feature

    writer = qdbd_connection.pinned_writer([table])
    _test_with_table(
        writer,
        table,
        many_intervals,
        _fast_push,
        _set_batch_writer_column_data)


def test_insert_column_truncate_implicit_range(
        qdbd_connection, table, many_intervals):

    whole_range = (
        many_intervals[0], many_intervals[-1:][0] + np.timedelta64(2, 's'))

    # Generate our dataset
    data = _generate_data(len(many_intervals))
    # (doubles, integers, blobs, strings, timestamps) = data
    (doubles, _, _, _, _) = data

    # Insert once
    writer = qdbd_connection.pinned_writer([table])
    _set_batch_writer_column_data(writer, many_intervals, data)
    writer.push()

    # Compare results, should be equal
    results = table.double_get_ranges(
        tslib._double_col_name(table), [whole_range])

    np.testing.assert_array_equal(results[0], many_intervals)
    np.testing.assert_array_equal(results[1], doubles)

    # Insert regular, twice
    _set_batch_writer_column_data(writer, many_intervals, data)
    writer.push()

    # Compare results, should now have the same data twice
    results = table.double_get_ranges(
        tslib._double_col_name(table), [whole_range])

    assert len(results[1]) == 2 * len(doubles)

    # Insert truncate, should now have original data again
    _set_batch_writer_column_data(writer, many_intervals, data)
    writer.push_truncate()

    # Verify results, truncating should now make things the same
    # as the beginning again.
    results = table.double_get_ranges(
        tslib._double_col_name(table), [whole_range])

    np.testing.assert_array_equal(results[0], many_intervals)
    np.testing.assert_array_equal(results[1], doubles)


def test_insert_column_truncate_explicit_range(
        qdbd_connection, table, many_intervals):

    whole_range = (
        many_intervals[0], many_intervals[-1:][0] + np.timedelta64(2, 's'))

    # Generate our dataset
    data = _generate_data(len(many_intervals))
    # (doubles, integers, blobs, strings, timestamps) = data
    (doubles, _, _, _, _) = data

    writer = qdbd_connection.pinned_writer([table])

    # Insert once
    truncate_range = (whole_range[0],
                      whole_range[1] + np.timedelta64(1, 'ns'))

    _set_batch_writer_column_data(writer, many_intervals, data)
    writer.push()

    # Verify results, truncating should now make things the same
    # as the beginning again.
    results = table.double_get_ranges(
        tslib._double_col_name(table), [whole_range])

    np.testing.assert_array_equal(results[0], many_intervals)
    np.testing.assert_array_equal(results[1], doubles)

    # If we now set the same data, skip the first element, but keep
    # the same time range, the first element will *not* be present in
    # the resulting dataset.
    _set_batch_writer_column_data(writer, many_intervals, data, start=1)
    writer.push_truncate(range=truncate_range)

    # Verify results, truncating should now make things the same
    # as the beginning again.
    results = table.double_get_ranges(
        tslib._double_col_name(table), [whole_range])

    np.testing.assert_array_equal(results[0], many_intervals[1:])
    np.testing.assert_array_equal(results[1], doubles[1:])


def test_insert_column_truncate_throws_error_on_invalid_range(
        qdbd_connection, table, many_intervals):
    whole_range = (
        many_intervals[0], many_intervals[-1:][0] + np.timedelta64(2, 's'))

    # Generate our dataset
    data = _generate_data(len(many_intervals))
    # (doubles, integers, blobs, strings, timestamps) = data
    (_, _, _, _, _) = data

    # Insert truncate with explicit timerange, we point the start right after the
    # first element in our dataset. This means that the range does not overlap all
    # the data anymore.
    truncate_range = (whole_range[0] + np.timedelta64(1, 'ns'),
                      whole_range[1] + np.timedelta64(1, 'ns'))

    writer = qdbd_connection.pinned_writer([table])
    _set_batch_writer_column_data(writer, many_intervals, data)
    with pytest.raises(quasardb.Error):
        writer.push_truncate(range=truncate_range)
